#!/usr/bin/env python3

import os
import torch
from biobb_pytorch.mdae.featurization.topology_selector import MDTopologySelector
from biobb_pytorch.mdae.featurization.featurizer import Featurizer
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger
import numpy as np
from typing import Optional, Dict, Any
from biobb_pytorch.mdae.utils.log_utils import get_size


class MDFeaturePipeline(BiobbObject):
    """
    | biobb_pytorch MDFeaturizer
    | Obtain the Molecular Dynamics Features for PyTorch model training.
    | Obtain the Molecular Dynamics Features for PyTorch model training.

    Args:
        input_trajectory_path (str) (Optional): Path to the input trajectory file (if omitted topology file is used as trajectory). File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/data/mdae/train_mdae_traj.xtc>`_. Accepted formats: xtc (edam:format_3875), dcd (edam:format_3878).
        input_topology_path (str): Path to the input topology file. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/data/mdae/MCV1900209.pdb>`_. Accepted formats: pdb (edam:format_2333).
        output_dataset_pt_path (str): Path to the output dataset model file. File type: output. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/ref_output_dataset.pt>`_. Accepted formats: pt (edam:format_2333).
        output_stats_pt_path (str): Path to the output model statistics file. File type: output. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/ref_output_stats.pt>`_. Accepted formats: pt (edam:format_2333).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **cartesian** (*dict*) - ({"selection": "name CA"}) Atom selection options for Cartesian coordinates feature generation (e.g. selection, fit_selection).
            * **distances** (*dict*) - ({"selection": "name CA", "cutoff": 0.4, "periodic": True, "bonded": False}) Atom selection options for pairwise distance features (selection, cutoff, periodic, bonded, etc.).
            * **angles** (*dict*) - ({"selection": "backbone", "periodic": True, "bonded": True}) Atom selection options for angle features (selection, periodic, bonded, etc.).
            * **dihedrals** (*dict*) - ({"selection": "backbone", "periodic": True, "bonded": True}) Atom selection options for dihedral features (selection, periodic, bonded, etc.).
            * **options** (*dict*) - ({"norm_in": {"mode": "min_max"}}) General processing options (e.g. timelag, norm_in).

    Examples:
        This is a use case of how to use the building block from Python::

            from biobb_pytorch.mdae.MDFeaturePipeline import MDFeaturizer

            prop = {
                'cartesian': {'selection': 'name CA'},
                'distances': {'selection': 'name CA',
                              'cutoff': 0.4,
                              'periodic': True,
                              'bonded': False},
                'angles': {'selection': 'backbone',
                           'periodic': True,
                           'bonded': True},
                'dihedrals': {'selection': 'backbone',
                              'periodic': True,
                              'bonded': True},
                'options': {'timelag': 10,
                            'norm_in': {'mode': 'min_max'}
                           }
            }

            MDFeaturizer(input_trajectory_path=trajectory_file,
                         input_topology_path=topology_file,
                         output_dataset_pt_path=output_file,
                         output_stats_pt_path=output_stats_file,
                         properties=prop)

    Info:
        * wrapped_software:
            * name: PyTorch
            * version: >=1.6.0
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
        self,
        input_topology_path: str,
        output_dataset_pt_path: str,
        output_stats_pt_path: str,
        properties: dict,
        input_trajectory_path: Optional[str] = None,
        input_labels_npy_path: Optional[str] = None,
        input_weights_npy_path: Optional[str] = None,
        **kwargs,
    ) -> None:

        properties = properties or {}

        super().__init__(properties)

        self.input_trajectory_path = input_trajectory_path or input_topology_path
        self.input_topology_path = input_topology_path
        self.input_labels_npy_path = input_labels_npy_path
        self.input_weights_npy_path = input_weights_npy_path
        self.output_dataset_pt_path = output_dataset_pt_path
        self.output_stats_pt_path = output_stats_pt_path
        self.config = properties.copy()
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_trajectory_path": input_trajectory_path,
                "input_topology_path": input_topology_path,
                "input_labels_npy_path": input_labels_npy_path,
                "input_weights_npy_path": input_weights_npy_path,
            },
            "out": {
                "output_dataset_pt_path": output_dataset_pt_path,
                "output_stats_pt_path": output_stats_pt_path,
            },
        }

        # build the per-feature arguments
        self.feature_types = ["cartesian", "distances", "angles", "dihedrals"]
        self.cartesian: dict = properties.get("cartesian", {"selection": "name CA"})
        self.distances: dict = properties.get("distances", {"selection": "name CA", "cutoff": 0.4, "periodic": True, "bonded": False})
        self.angles: dict = properties.get("angles", {"selection": "backbone", "periodic": True, "bonded": True})
        self.dihedrals: dict = properties.get("dihedrals", {"selection": "backbone", "periodic": True, "bonded": True})
        self.options: dict = properties.get("options", {"norm_in": {"mode": "min_max"}})

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

        # Topology indices
        self.topology_indices()

        # Featurizer
        self.featurize_trajectory()

    @launchlogger
    def topology_indices(self) -> Dict[str, Any]:

        fu.log("## BioBB Featurization - MDFeaturePipeline ##", self.out_log)

        fu.log(f"Obtaining the topology information from {self.input_topology_path}", self.out_log)

        self.topology = MDTopologySelector(self.input_topology_path)
        self.features_idx_dict = self.topology.topology_indexing(self.config)

        fu.log("Available Topology Properties:", self.out_log)
        fu.log(f"  - Number of chains: {self.topology.topology.n_chains}", self.out_log)
        fu.log(f"  - Number of residues: {self.topology.topology.n_residues}", self.out_log)
        fu.log(f"  - Number of atoms: {self.topology.n_atoms}", self.out_log)
        try:
            fu.log(f"  - Number of distances: {self.topology.n_distances}", self.out_log)
        except AttributeError:
            fu.log("  - Number of distances: N/A", self.out_log)
        try:
            fu.log(f"  - Number of angles: {self.topology.n_angles}", self.out_log)
        except AttributeError:
            fu.log("  - Number of angles: N/A", self.out_log)
        try:
            fu.log(f"  - Number of dihedrals: {self.topology.n_dihedrals}", self.out_log)
        except AttributeError:
            fu.log("  - Number of dihedrals: N/A", self.out_log)

    @launchlogger
    def featurize_trajectory(self) -> None:

        self.featurizer = Featurizer(self.input_trajectory_path,
                                     self.input_topology_path,
                                     self.input_labels_npy_path,
                                     self.input_weights_npy_path,
                                     )

        fu.log("Available Trajectory Properties:", self.out_log)
        fu.log(f"   - Number of frames: {self.featurizer.trajectory.n_frames}", self.out_log)

        fu.log(f"Featurizing the trajectory {self.input_trajectory_path}", self.out_log)

        self.dataset, self.stats = self.featurizer.compute_features(self.features_idx_dict)

        if self.input_labels_npy_path:
            fu.log(f"Loading labels from {self.input_labels_npy_path}", self.out_log)
            self.dataset['labels'] = np.load(self.input_labels_npy_path)

        if self.input_weights_npy_path:
            fu.log(f"Loading weights from {self.input_weights_npy_path}", self.out_log)
            self.dataset['weights'] = np.load(self.input_weights_npy_path)

        fu.log("Features:", self.out_log)
        for feature_type in self.feature_types:
            try:
                selection = getattr(self, feature_type).get("selection")
                shape = self.featurizer.features.get(feature_type, np.zeros((0, 0))).shape[1]
                fu.log(f"  {feature_type.capitalize()}:", self.out_log)
                fu.log(f"   - Topology Selection: {selection}", self.out_log)
                fu.log(f"   - Number of features: {shape}", self.out_log)
            except AttributeError:
                pass

        fu.log("Postprocessing:", self.out_log)
        fu.log(f"   - Normalization: {self.options.get('norm_in', {}).get('mode')}", self.out_log)
        fu.log(f"   - Timelag: {self.options.get('timelag', {})}", self.out_log)
        fu.log("Dataset Properties:", self.out_log)
        fu.log(f"   - Dataset: {self.dataset.keys()}", self.out_log)
        fu.log(f"   - Number of frames: {self.dataset['data'].shape[0]}", self.out_log)
        fu.log(f"   - Number of features: {self.dataset['data'].shape[1]}", self.out_log)

    @launchlogger
    def launch(self) -> int:
        """
        Execute the :class:`MDFeaturePipeline <MDFeaturePipeline.MDFeaturePipeline>` object
        """

        # Setup Biobb
        if self.check_restart():
            return 0

        self.stage_files()

        torch.save(self.dataset,
                   self.output_dataset_pt_path)

        fu.log(f'Dataset saved in .pt format in {os.path.abspath(self.io_dict["out"]["output_dataset_pt_path"])}',
               self.out_log,
               )
        fu.log(f'File size: {get_size(self.io_dict["out"]["output_dataset_pt_path"])}',
               self.out_log,
               )

        torch.save(self.stats,
                   os.path.splitext(self.output_stats_pt_path)[0] + ".pt")

        fu.log(f'Dataset statistics saved in .pt format in {os.path.abspath(self.io_dict["out"]["output_stats_pt_path"])}',
               self.out_log,
               )
        fu.log(f'File size: {get_size(self.io_dict["out"]["output_stats_pt_path"])}',
               self.out_log,
               )

        # Copy files to host
        self.copy_to_host()

        # Remove temporal files
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0


def MDFeaturizer(
    input_topology_path: str,
    output_dataset_pt_path: str,
    output_stats_pt_path: str,
    properties: dict,
    input_trajectory_path: Optional[str] = None,
    input_labels_npy_path: Optional[str] = None,
    input_weights_npy_path: Optional[str] = None,
    **kwargs,
) -> int:
    """Create the :class:`MDFeaturePipeline <MDFeaturePipeline.MDFeaturePipeline>` class and
    execute the :meth:`launch() <MDFeaturePipeline.MDFeaturizer.launch>` method."""
    return MDFeaturePipeline(**dict(locals())).launch()


MDFeaturizer.__doc__ = MDFeaturePipeline.__doc__
main = MDFeaturePipeline.get_main(MDFeaturizer, "Obtain the Molecular Dynamics Features for PyTorch model training.")

if __name__ == "__main__":
    main()



from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger
import torch
import numpy as np
import mdtraj as md
import os


class Feat2Traj(BiobbObject):
    """
    | biobb_pytorch Feat2Traj
    | Converts a .pt file (features) to a trajectory using cartesian indices and topology from the stats file.
    | Converts a .pt file (features) to a trajectory using cartesian indices and topology from the stats file.

    Args:
        input_results_npz_path (str): Path to the input reconstructed results file (.npz), typically containing an 'xhat' array. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/ref_input_results.npz>`_. Accepted formats: npz (edam:format_2333).
        input_stats_pt_path (str): Path to the input model statistics file (.pt) containing cartesian indices and optionally topology. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/ref_input_model.pt>`_. Accepted formats: pt (edam:format_2333).
        input_topology_path (str) (optional): Path to the topology file (PDB) used if no suitable topology is found in the stats file. Used if no topology is found in stats. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/mdae/ref_input_topology.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_traj_path (str): Path to save the trajectory in xtc/pdb/dcd format. File type: output. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/output_model.xtc>`_. Accepted formats: xtc (edam:format_3875), pdb (edam:format_1476), dcd (edam:format_3878).
        output_top_path (str) (optional): Path to save the output topology file (pdb). Used if trajectory format requires separate topology. File type: output. `Sample file <https://github.com/bioexcel/biobb_pytorch/mdae/output_model.pdb>`_. Accepted formats: pdb (edam:format_1476).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.

    Examples:
        This example shows how to use the Feat2Traj class to convert a .pt file (features) to a trajectory using cartesian indices and topology from the stats file::

            from biobb_pytorch.mdae.feat2traj import feat2traj

            input_results_npz_path='input_results.npz'
            input_stats_pt_path='input_model.pt'
            input_topology_path='input_topology.pdb'
            output_traj_path='output_model.xtc'
            output_top_path='output_model.pdb'

            prop={}

            feat2traj(input_results_npz_path=input_results_npz_path,
                    input_stats_pt_path=input_stats_pt_path,
                    input_topology_path=input_topology_path,
                    output_traj_path=output_traj_path,
                    output_top_path=output_top_path,
                    properties=prop)

    Info:
        * wrapped_software:
            * name: PyTorch
            * version: >=1.6.0
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
        self,
        input_results_npz_path: str,
        input_stats_pt_path: str,
        input_topology_path: str = None,
        output_traj_path: str = None,
        output_top_path: str = None,
        properties: dict = None,
        **kwargs,
    ) -> None:
        properties = properties or {}
        super().__init__(properties)

        self.input_results_npz_path = input_results_npz_path
        self.input_stats_pt_path = input_stats_pt_path
        self.input_topology_path = input_topology_path
        self.output_traj_path = output_traj_path
        self.output_top_path = output_top_path
        self.properties = properties.copy()
        self.locals_var_dict = locals().copy()
        self.io_dict = {
            "in": {
                "input_results_npz_path": input_results_npz_path,
                "input_stats_pt_path": input_stats_pt_path,
                "input_topology_path": input_topology_path,
            },
            "out": {
                "output_traj_path": output_traj_path,
                "output_top_path": output_top_path,
            },
        }
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """
        Execute the :class:`Feat2Traj` class and its `.launch()` method.
        """
        # Load features
        features = np.load(self.input_results_npz_path)
        features = features['xhat']

        # Load stats and extract cartesian indices and topology
        stats = torch.load(self.input_stats_pt_path,
                           weights_only=False)
        cartesian_indices = None
        topology = None
        if isinstance(stats, dict):
            if 'cartesian_indices' in stats:
                cartesian_indices = stats['cartesian_indices']
                topology = stats['topology']

        else:
            raise ValueError('No cartesian indices found in stats file.')
        cartesian_indices = np.array(cartesian_indices)

        n_atoms = len(cartesian_indices)
        n_frames = features.shape[0]
        coords = features.reshape((n_frames, n_atoms, 3))

        # Try to use topology from stats file if present
        top = None
        if topology is not None:
            try:
                # If topology is a serialized MDTraj Topology, try to load it
                if isinstance(topology, md.Trajectory):
                    top = topology.topology
                elif isinstance(topology, str) and os.path.exists(topology):
                    top = md.load_topology(topology)
                elif isinstance(topology, dict) and 'pdb_string' in topology:
                    import io
                    top = md.load(io.StringIO(topology['pdb_string']), format='pdb').topology
            except Exception as e:
                print(f"Warning: Could not load topology from stats file: {e}")
                top = None

        # If not found, try input_topology_path
        if top is None and self.input_topology_path is not None and os.path.exists(self.input_topology_path):
            top = md.load_topology(self.input_topology_path)
        # Fallback: create a fake topology
        if top is None:
            top = md.Topology()
            chain = top.add_chain()
            res = top.add_residue('RES', chain)
            for i in range(n_atoms):
                top.add_atom('CA', element=md.element.carbon, residue=res)
        traj = md.Trajectory(xyz=coords, topology=top)

        if self.output_traj_path:
            ext = os.path.splitext(self.output_traj_path)[1]
            if ext == '.xtc':
                traj.save_xtc(self.output_traj_path)
                traj[0].save_pdb(self.output_top_path)
            elif ext == '.dcd':
                traj.save_dcd(self.output_traj_path)
                traj[0].save_pdb(self.output_top_path)
            elif ext == '.pdb':
                traj.save_pdb(self.output_traj_path)
            else:
                raise ValueError(f'Unknown trajectory extension: {ext}')
        return 0


def feat2traj(
    input_results_npz_path: str,
    input_stats_pt_path: str,
    input_topology_path: str = None,
    output_traj_path: str = None,
    output_top_path: str = None,
    properties: dict = None,
    **kwargs,
) -> int:
    """Create the :class:`Feat2Traj <Feat2Traj.Feat2Traj>` class and
    execute the :meth:`launch() <Feat2Traj.feat2traj.launch>` method."""
    return Feat2Traj(**dict(locals())).launch()


feat2traj.__doc__ = Feat2Traj.__doc__
main = Feat2Traj.get_main(feat2traj, "Converts a .pt file (features) to a trajectory using cartesian indices and topology from the stats file.")

if __name__ == "__main__":
    main()



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Reduce eigenvalues loss.
"""

__all__ = ["ReduceEigenvaluesLoss", "reduce_eigenvalues_loss"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import torch


# =============================================================================
# LOSS FUNCTIONS
# =============================================================================


class ReduceEigenvaluesLoss(torch.nn.Module):
    """Calculate a monotonic function f(x) of the eigenvalues, by default the sum.

    By default it returns -f(x) to be used as loss function to maximize
    eigenvalues in gradient descent schemes.

    The following reduce functions are implemented:
        - sum     : sum_i (lambda_i)
        - sum2    : sum_i (lambda_i)**2
        - gap     : (lambda_1-lambda_2)
        - its     : sum_i (1/log(lambda_i))
        - single  : (lambda_i)
        - single2 : (lambda_i)**2

    """

    def __init__(
        self,
        mode: str = "sum",
        n_eig: int = 0,
        invert_sign: bool = True,
    ):
        """Constructor.

        Parameters
        ----------
        mode : str, optional
            Function of the eigenvalues to optimize (see notes). Default is ``'sum'``.
        n_eig: int, optional
            Number of eigenvalues to include in the loss (default: 0 --> all).
            In case of ``'single'`` and ``'single2'`` is used to specify which
            eigenvalue to use.
        invert_sign: bool, optional
            Whether to return the opposite of the function (in order to be minimized
            with GD methods). Default is ``True``.
        """
        super().__init__()
        self.mode = mode
        self.n_eig = n_eig
        self.invert_sign = invert_sign

    def forward(self, evals: torch.Tensor) -> torch.Tensor:
        """Compute the loss.

        Parameters
        ----------
        evals : torch.Tensor
            Shape ``(n_batches, n_eigenvalues)``. Eigenvalues.

        Returns
        -------
        loss : torch.Tensor
        """
        return reduce_eigenvalues_loss(evals, self.mode, self.n_eig, self.invert_sign)


def reduce_eigenvalues_loss(
    evals: torch.Tensor,
    mode: str = "sum",
    n_eig: int = 0,
    invert_sign: bool = True,
) -> torch.Tensor:
    """Calculate a monotonic function f(x) of the eigenvalues, by default the sum.

    By default it returns -f(x) to be used as loss function to maximize
    eigenvalues in gradient descent schemes.

    Parameters
    ----------
    evals : torch.Tensor
        Shape ``(n_batches, n_eigenvalues)``. Eigenvalues.
    mode : str, optional
        Function of the eigenvalues to optimize (see notes). Default is ``'sum'``.
    n_eig: int, optional
        Number of eigenvalues to include in the loss (default: 0 --> all).
        In case of ``'single'`` and ``'single2'`` is used to specify which
        eigenvalue to use.
    invert_sign: bool, optional
        Whether to return the opposite of the function (in order to be minimized
        with GD methods). Default is ``True``.

    Notes
    -----
    The following functions are implemented:
        - sum     : sum_i (lambda_i)
        - sum2    : sum_i (lambda_i)**2
        - gap     : (lambda_1-lambda_2)
        - its     : sum_i (1/log(lambda_i))
        - single  : (lambda_i)
        - single2 : (lambda_i)**2

    Returns
    -------
    loss : torch.Tensor (scalar)
        Loss value.
    """

    # check if n_eig is given and
    if (n_eig > 0) & (len(evals) < n_eig):
        raise ValueError("n_eig must be lower than the number of eigenvalues.")
    elif n_eig == 0:
        if (mode == "single") | (mode == "single2"):
            raise ValueError("n_eig must be specified when using single or single2.")
        else:
            n_eig = len(evals)

    loss = None

    if mode == "sum":
        loss = torch.sum(evals[:n_eig])
    elif mode == "sum2":
        g_lambda = torch.pow(evals, 2)
        loss = torch.sum(g_lambda[:n_eig])
    elif mode == "gap":
        loss = evals[0] - evals[1]
    elif mode == "its":
        g_lambda = 1 / torch.log(evals)
        loss = torch.sum(g_lambda[:n_eig])
    elif mode == "single":
        loss = evals[n_eig - 1]
    elif mode == "single2":
        loss = torch.pow(evals[n_eig - 1], 2)
    else:
        raise ValueError(
            f"unknown mode : {mode}. options: 'sum','sum2','gap','single','its'."
        )

    if invert_sign:
        loss *= -1

    return loss



import torch
import torch.nn as nn
import torch.nn.functional as F


def lrp_gmvae_single(model, x, latent_index, eps=1e-6):
    device = x.device
    batch_size = x.shape[0]
    in_features = model.in_features
    k = model.k

    has_norm = model.norm_in is not None
    if has_norm:
        x_input = model.norm_in(x)
    else:
        x_input = x

    qy_logit = model.encoder['qy_nn'](x_input)
    qy = torch.softmax(qy_logit, dim=1)

    y_ = torch.zeros(batch_size, k, device=device)

    zm_list = []
    intermediates_list = []

    for i in range(k):
        y = y_ + torch.eye(k, device=device)[i]

        # y_transform
        module = model.encoder['y_transform']
        z_h0 = y @ module.weight.t() + module.bias
        a_h0 = z_h0  # no activation

        intermediates = [('linear', module, y, z_h0)]

        xy = torch.cat([x_input, a_h0], dim=1)

        a = xy

        # qz_nn
        for sub_module in model.encoder['qz_nn'].children():
            if isinstance(sub_module, nn.Linear):
                z = a @ sub_module.weight.t() + sub_module.bias
                intermediates.append(('linear', sub_module, a, z))
                a = z
            elif isinstance(sub_module, nn.ReLU):
                a = F.relu(a)

        # zm_layer
        module = model.encoder['zm_layer']
        z = a @ module.weight.t() + module.bias
        intermediates.append(('linear', module, a, z))

        zm = z
        zm_list.append(zm)

        intermediates_list.append(intermediates)

    zm = torch.stack(zm_list, dim=1)  # [batch, k, n_cvs]

    term_k = qy * zm[:, :, latent_index]  # [batch, k]

    selected_a = torch.sum(term_k, dim=1)  # [batch]

    R_a = selected_a.unsqueeze(1)  # [batch,1]

    sign_a = selected_a.sign().unsqueeze(1)

    den = selected_a.unsqueeze(1) + eps * sign_a

    R_term_k = R_a * (term_k.unsqueeze(2) / den)  # [batch, k,1]

    R0 = torch.zeros(batch_size, in_features, device=device)

    for i in range(k):
        R = R_term_k[:, i, :]  # [batch,1]

        intermediates = intermediates_list[i][::-1]

        for op, module, a_prev, z_prev in intermediates:
            if op == 'linear':
                sign_z = z_prev.sign()
                Z = z_prev + eps * sign_z
                s = R / Z
                c = s @ module.weight
                R = a_prev * c

        # R is now R for xy
        R_xy = R
        R_x_k = R_xy[:, :in_features]
        R0 += R_x_k

    if has_norm:
        w = (1 / model.norm_in.range).view(1, -1).to(device)
        b = (-model.norm_in.mean / model.norm_in.range).view(1, -1).to(device)
        z = x * w + b
        sign_z = z.sign()
        Z = z + eps * sign_z
        s = R0 / Z
        c = s * w
        R0 = x * c

    return R0


def lrp_encoder(
    model: nn.Module,
    x: torch.Tensor,
    latent_index: int = None,
    eps: float = 1e-6
) -> torch.Tensor:
    """Main LRP function that handles both full models and encoder modules."""
    # Check if it's a GMVAE model
    if model.__class__.__name__ == 'GaussianMixtureVariationalAutoEncoder':
        if latent_index is None:
            R0 = 0
            for j in range(model.out_features):
                R0 += lrp_gmvae_single(model, x, j, eps)
            return R0
        else:
            return lrp_gmvae_single(model, x, latent_index, eps)

    # Check if model has forward_cv method (full model with encoder)
    if hasattr(model, 'forward_cv'):
        # General case for MLP models with forward_cv
        handles = []
        layers = []

        def collect_hook(module, inp, out):
            layers.append(module)

        for m in model.modules():
            if isinstance(m, nn.Linear):
                handle = m.register_forward_hook(collect_hook)
                handles.append(handle)

        model.eval()
        with torch.no_grad():
            _ = model.forward_cv(x[:1])

        for h in handles:
            h.remove()

        if len(layers) == 0:
            raise ValueError("No Linear layers found.")

        for layer in layers:
            if not isinstance(layer, nn.Linear):
                raise ValueError("LRP only supported for Linear layers in general case.")

        L = len(layers)

        has_norm = hasattr(model, 'norm_in') and model.norm_in is not None
        if has_norm:
            input_to_encoder = model.norm_in(x)
        else:
            input_to_encoder = x

        A = [input_to_encoder.clone()]
        Z = [None] * (L + 1)
        for layer_idx in range(L):
            lin = layers[layer_idx]
            z = A[layer_idx] @ lin.weight.t() + lin.bias
            sign_z = z.sign()
            Z[layer_idx + 1] = z + eps * sign_z
            if layer_idx < L - 1:
                a = F.relu(z)
            else:
                a = z
            A.append(a)

        zL = A[L]
        if latent_index is None:
            R = [None] * (L + 1)
            R[L] = zL.sum(dim=1, keepdim=True)
        else:
            R = [None] * (L + 1)
            R[L] = zL[:, [latent_index]]

        for layer_idx in range(L - 1, -1, -1):
            lin = layers[layer_idx]
            s = R[layer_idx + 1] / Z[layer_idx + 1]
            c = s @ lin.weight
            R[layer_idx] = A[layer_idx] * c

        R0 = R[0]

        if has_norm:
            w = (1 / model.norm_in.range).view(1, -1).to(x.device)
            b = (-model.norm_in.mean / model.norm_in.range).view(1, -1).to(x.device)
            z = x * w + b
            sign_z = z.sign()
            Z = z + eps * sign_z
            s = R0 / Z
            c = s * w
            R0 = x * c

        return R0

    # Fall back to simple encoder version
    return _lrp_encoder_simple(model, x, latent_index, eps)


# Layer-wise Relevance Propagation (simplified version for encoder modules)
def _lrp_encoder_simple(
    encoder: nn.Module,
    x: torch.Tensor,
    latent_index: int = None,
    eps: float = 1e-6
) -> torch.Tensor:
    """
    Perform Layer‐Wise Relevance Propagation on `encoder` for input `x`.

    Arguments:
      encoder       -- an nn.Module mapping [batch, in_dim] → [batch, latent_dim],
                       built from Linear + ReLU layers.
      x             -- input tensor, shape [batch, in_dim].
      latent_index  -- which coordinate of the latent vector to explain.
                       If None, we explain the sum over all latent dims.
      eps           -- stabilization term to avoid division by zero.

    Returns:
      R0            -- relevance at the input layer, shape [batch, in_dim].
                       R0[b, i] is “how important feature i was for the chosen
                       latent coordinate (or sum).”
    """
    device = x.device

    # 1) Extract all Linear layers in execution order
    layers = []
    for module in encoder.modules():
        if isinstance(module, nn.Linear):
            layers.append(module)
    L = len(layers)

    # 2) FORWARD PASS: collect activations A[l] and pre-activations Z[l]
    A = [x.clone().to(device)]
    Z = [None] * (L + 1)
    for layer_idx, lin in enumerate(layers):
        z = A[layer_idx] @ lin.weight.t() + lin.bias  # shape [batch, out_dim]
        Z[layer_idx + 1] = z + eps                       # add eps for numerical stability
        a = F.relu(z)
        A.append(a)

    # 3) INITIALIZE RELEVANCE at the top
    #    Let zL = A[L] be shape [batch, latent_dim]
    zL = A[L]
    if latent_index is None:
        # explain the sum of all latent coords
        R = [None] * (L + 1)
        R[L] = zL.sum(dim=1, keepdim=True)    # shape [batch, 1]
    else:
        # explain a single latent coordinate
        R = [None] * (L + 1)
        R[L] = zL[:, [latent_index]]          # shape [batch, 1]

    # 4) BACKWARD PASS (LRP) from layer L → 0
    #    At each step layer_idx, we have R[layer_idx+1] of shape [batch, out_dim].
    #    We want R[layer_idx] of shape [batch, in_dim].
    for layer_idx in range(L - 1, -1, -1):
        lin = layers[layer_idx]
        w = lin.weight       # shape [out_dim, in_dim]
        # Z[layer_idx+1]: [batch, out_dim], R[layer_idx+1]: [batch, out_dim]
        s = R[layer_idx + 1] / Z[layer_idx + 1]               # [batch, out_dim]
        c = s @ w                         # [batch, in_dim], since w is [out, in]
        # multiply by the forward activation to zero‐out inactive neurons
        R[layer_idx] = A[layer_idx] * c                   # [batch, in_dim]

    # R[0] is now the relevance of each input feature
    return R[0]

# device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# X_batch1 = torch.tensor(data[2000:4000]).to(device)

# # explain the sum of *all* latent coords:
# R0_sum = lrp_encoder(model.encoder, X_batch1, latent_index=None)

# # To get a global feature ranking, average absolute relevance over the batch:
# R0_sum = R0_sum.reshape(R0_sum.size(0), -1, 3)

# R0_sum = R0_sum.mean(dim=2)     # [in_dim]
# global_importance = R0_sum.abs().mean(dim=0)    # [in_dim]
# global_importance = global_importance.cpu().detach().numpy()

# # Normalize
# global_importance = (global_importance - global_importance.min()) / (global_importance.max() - global_importance.min())



# Copyright (c) 2021 Venkata K. Ramaswamy, Samuel C. Musson, Chris G. Willcocks, Matteo T. Degiacomi
#
# Molearn is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# molearn is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with molearn ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

import numpy as np
import torch
from copy import deepcopy
import os
from importlib.resources import files


def read_lib_file(file_name, amber_atoms, atom_charge, connectivity):
    try:
        f_location = files('mdae').joinpath('loss/parameters')
        f_in = open(f'{str(f_location)}/{file_name}')
        print(f'File {f_location}/{file_name} opened')
    except Exception:
        raise Exception(f'ERROR: file {file_name} not found!')

    lines = f_in.readlines()
    depth = 0
    indexs = {}
    for tline in lines:
        if tline.split() == ['!!index', 'array', 'str']:
            depth += 1
            for line in lines[depth:]:
                if line[0] != ' ':
                    break
                contents = line.split()
                if len(contents) != 1 and len(contents[0]) != 5:
                    break
                res = contents[0]
                if res[0] == '"' and res[-1] == '"':
                    amber_atoms[res[1:-1]] = {}
                    atom_charge[res[1:-1]] = {}
                    indexs[res[1:-1]] = {}
                    connectivity[res[1:-1]] = {}
                else:
                    msg = 'I was expecting something of the form "XXX" but got %s instead' % res
                    raise Exception(msg)
                depth += 1
            break
        depth += 1

    for i, tline in enumerate(lines):
        entry, res, unit_atoms, unit_connectivity = tline[0:7], tline[7:10], tline[10:22], tline[10:29]
        if entry == '!entry.' and unit_atoms == '.unit.atoms ':
            depth = i + 1
            for line in lines[depth:]:
                if line[0] != ' ':
                    break
                contents = line.split()
                if len(contents) < 3 and len(contents[0]) > 4 and len(contents[1]) > 4:
                    break
                pdb_name, amber_name, _, _, _, index, element_number, charge = contents
                pdb_quoted = pdb_name[0] == '"' and pdb_name[-1] == '"'
                amber_quoted = amber_name[0] == '"' and amber_name[-1] == '"'
                if pdb_quoted and amber_quoted:
                    amber_atoms[res][contents[0][1:-1]] = contents[1][1:-1]
                    atom_charge[res][amber_name[1:-1]] = float(charge)
                    # indexs[res][amber_name[1:-1]] = int(index)
                    indexs[res][int(index)] = pdb_name[1:-1]
                    connectivity[res][pdb_name[1:-1]] = []
                else:
                    msg = 'I was expecting something of the form "XXX" but got %s instead' % res
                    raise Exception(msg)
        elif entry == '!entry.' and unit_connectivity == '.unit.connectivity ':
            depth = i + 1
            for line in lines[depth:]:
                if line[0] != ' ':
                    break
                contents = line.split()
                if len(contents) != 3:
                    break
                a1, a2, flag = contents
                connectivity[res][indexs[res][int(a1)]].append(indexs[res][int(a2)])
                connectivity[res][indexs[res][int(a2)]].append(indexs[res][int(a1)])


def get_amber_parameters(order=False, radians=True):

    file_names = ('amino12.lib',
                  'parm10.dat',
                  'frcmod.ff14SB')

    # amber19 is dangerous because they've replaced parameters with cmap
    # pdb atom names to amber atom names using amino19.lib
    amber_atoms = {}  # knowledge[res][pdb_atom] = amber_atom
    atom_mass = {}
    atom_polarizability = {}
    bond_force = {}
    bond_equil = {}
    angle_force = {}
    angle_equil = {}
    torsion_factor = {}
    torsion_barrier = {}
    torsion_phase = {}
    torsion_period = {}
    improper_factor = {}
    improper_barrier = {}
    improper_phase = {}
    improper_period = {}

    other_parameters = {}
    other_parameters['vdw_potential_well_depth'] = {}
    other_parameters['H_bond_10_12_parameters'] = {}
    other_parameters['equivalences'] = {}
    other_parameters['charge'] = {}
    other_parameters['connectivity'] = {}
    read_lib_file(file_names[0], amber_atoms, other_parameters['charge'], other_parameters['connectivity'])

    try:
        f_location = files('mdae').joinpath('loss/parameters')
        f_in = open(f'{str(f_location)}/{file_names[1]}')
        print('File %s opened' % file_names[1])
    except Exception:
        raise Exception('ERROR: file %s not found!' % file_names[1])

    # section 1 title
    line = f_in.readline()
    print(line)

    amber_card_type_2(f_in, atom_mass, atom_polarizability)
    amber_card_type_3(f_in)
    amber_card_type_4(f_in, bond_force, bond_equil)
    amber_card_type_5(f_in, angle_force, angle_equil)
    amber_card_type_6(f_in, torsion_factor, torsion_barrier, torsion_phase, torsion_period)
    amber_card_type_7(f_in, improper_factor,
                      improper_barrier, improper_phase, improper_period)
    amber_card_type_8(f_in, other_parameters)
    amber_card_type_9(f_in, other_parameters)
    for line in f_in:
        if len(line.split()) > 1:
            if line.split()[1] == 'RE':
                amber_card_type_10B(f_in, other_parameters)
        elif line[0:3] == 'END':
            print('parameters loaded')
    f_in.close()

    # pen frcmod file, should be identifcal format but missing any or all cards
    try:
        f_location = files('mdae').joinpath('loss/parameters')
        f_in = open(f'{str(f_location)}/{file_names[2]}')
        print('File %s opened' % file_names[2])
    except Exception:
        raise Exception('ERROR: file %s not found!' % file_names[2])

    # section 1 title
    line = f_in.readline()
    print(line)

    for line in f_in:
        if line[:4] == 'MASS':
            amber_card_type_2(f_in, atom_mass, atom_polarizability)
        if line[:4] == 'BOND':
            amber_card_type_4(f_in, bond_force, bond_equil)
        if line[:4] == 'ANGL':
            amber_card_type_5(f_in, angle_force, angle_equil)
        if line[:4] == 'DIHE':
            amber_card_type_6(f_in, torsion_factor, torsion_barrier, torsion_phase, torsion_period)
        if line[:4] == 'IMPR':
            amber_card_type_7(f_in, improper_factor,
                              improper_barrier, improper_phase, improper_period)
        if line[:4] == 'HBON':
            amber_card_type_8(f_in, other_parameters)
        if line[:4] == 'NONB':
            amber_card_type_10B(f_in, other_parameters)
        if line[:4] == 'CMAP':
            print('Yeah, Im not bothering to implement cmap')
        elif line[0:3] == 'END':
            print('parameters loaded')
    f_in.close()

    if radians:
        for angle in angle_equil:
            angle_equil[angle] = np.deg2rad(angle_equil[angle])
        for torsion in torsion_phase:
            torsion_phase[torsion] = list(np.deg2rad(torsion_phase[torsion]))

    return (amber_atoms, atom_mass, atom_polarizability, bond_force, bond_equil,
            angle_force, angle_equil, torsion_factor, torsion_barrier, torsion_phase,
            torsion_period, improper_factor, improper_barrier, improper_phase,
            improper_period, other_parameters)


def amber_card_type_2(f_in, atom_mass, atom_polarizability):

    # section 2 input for atom symbols and masses
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        atom = line[0:2].strip()
        contents = line[2:24].split()
        if len(contents) == 2:
            mass, polarizability = float(contents[0]), float(contents[1])
            atom_mass[atom] = mass
            atom_polarizability[atom] = polarizability
        elif len(contents) == 1:  # sometimes a polarizability is not listed
            mass = float(contents[0])
            atom_mass[atom] = mass
            atom_polarizability[atom] = polarizability
        else:
            raise Exception('Should be 2A, X, F10.2, F10.2, comments but got %s' % line)


def amber_card_type_3(f_in):
    # section 3  input for atom symbols that are hydrophilic
    f_in.readline()


def amber_card_type_4(f_in, bond_force, bond_equil, order=False):
    # section 4  bond length paramters
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        atom1 = line[0:2]. strip()
        atom2 = line[3:5].strip()
        if order:
            bond = tuple(sorted((atom1, atom2)))  # put in alphabetical order
        else:
            bond = (atom1, atom2)
        contents = line[5:25].split()
        if len(contents) != 2:
            raise Exception('Expected 2 floats but got %s' % line[6:26])
        force_constant, equil_length = float(contents[0]), float(contents[1])
        bond_force[bond] = force_constant
        bond_equil[bond] = equil_length
        # this should throw an error if there are not


def amber_card_type_5(f_in, angle_force, angle_equil, order=False):
    # section 5
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        atom1 = line[0:2].strip()
        atom2 = line[3:5].strip()
        atom3 = line[6:8].strip()
        if order:
            sorted13 = sorted((atom1, atom3))
            angle = (sorted13[0], atom2, sorted13[1])
            # I want it sorted alphabetically by 1-3 atoms
        else:
            angle = (atom1, atom2, atom3)
        contents = line[8:28].split()
        if len(contents) != 2:
            raise Exception('Expected 2 floats but got %s' % line[6:26])
        force_constant, equil_angle = float(contents[0]), float(contents[1])
        angle_force[angle] = force_constant
        angle_equil[angle] = equil_angle


def amber_card_type_6(f_in, torsion_factor, torsion_barrier, torsion_phase,
                      torsion_period, order=False):
    # section 6 torsion / proper dihedral
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        atom1 = line[0:2].strip()
        atom2 = line[3:5].strip()
        atom3 = line[6:8].strip()
        atom4 = line[9:11].strip()
        if order:
            sort23 = sorted([(atom2, atom1), (atom3, atom4)], key=lambda x: x[0])
            torsion = tuple((sort23[0][1], sort23[0][0], sort23[1][0], sort23[1][1]))
        else:
            torsion = (atom1, atom2, atom3, atom4)
        contents = line[11:55].split()
        if len(contents) != 4:
            raise Exception('I wanted four values here?')
        # the actual torsion potential is (barrier/factor)*(1+cos(period*phi-phase))
        if torsion in torsion_period:
            if torsion_period[torsion][-1] > 0:
                torsion_factor[torsion] = [int(contents[0])]
                torsion_barrier[torsion] = [float(contents[1])]
                torsion_phase[torsion] = [float(contents[2])]
                torsion_period[torsion] = [float(contents[3])]
            elif torsion_period[torsion][-1] < 0:
                torsion_factor[torsion].append(int(contents[0]))
                torsion_barrier[torsion].append(float(contents[1]))
                torsion_phase[torsion].append(float(contents[2]))
                torsion_period[torsion].append(float(contents[3]))
        else:
            torsion_factor[torsion] = [int(contents[0])]
            torsion_barrier[torsion] = [float(contents[1])]
            torsion_phase[torsion] = [float(contents[2])]
            torsion_period[torsion] = [float(contents[3])]


def amber_card_type_7(f_in, improper_factor, improper_barrier,
                      improper_phase, improper_period, order=False):
    # section 7 improper dihedrals
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        atom1 = line[0:2].strip()
        atom2 = line[3:5].strip()
        atom3 = line[6:8].strip()
        atom4 = line[9:11].strip()
        if order:
            sort23 = sorted([(atom2, atom1), (atom3, atom4)], key=lambda x: x[0])
            torsion = tuple((sort23[0][1], sort23[0][0], sort23[1][0], sort23[1][1]))
        else:
            torsion = (atom1, atom2, atom3, atom4)
        contents = line[11:55].split()
        if len(contents) == 3:
            improper_barrier[torsion] = float(contents[0])
            improper_phase[torsion] = float(contents[1])
            improper_period[torsion] = float(contents[2])
        elif len(contents) == 4:
            raise Exception('This seems allowed in the doc but doesnt appear in reality')
            improper_factor[torsion] = int(contents[0])
            improper_barrier[torsion] = float(contents[1])
            improper_phase[torsion] = float(contents[2])
            improper_period[torsion] = float(contents[3])
        # the actual torsion potential is (barrier/factor)*(1+cos(period*phi-phase))
        # it seems improper potential don't divide by the factor


def amber_card_type_8(f_in, other_parameters, order=False):
    # section 8  H-bond 10-12 potential parameters
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        atom1 = line[2:4].strip()
        atom2 = line[6:8].strip()
        if order:
            pair = tuple(sorted((atom1, atom2)))
        else:
            pair = (atom1, atom2)
        contents = line[8:].split()
        other_parameters['H_bond_10_12_parameters'][pair] = contents


def amber_card_type_9(f_in, other_parameters):
    # section 9 equivalencing atom symbols for non-bonded 6-12 potential parameters
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        contents = line.split()
        other_parameters['equivalences'][contents[0]] = contents


def amber_card_type_10B(f_in, other_parameters):
    # section 10 6-12  potential parameters
    for line in f_in:
        if line == '\n' or line.strip() == '':
            break
        contents = line.split()
        other_parameters['vdw_potential_well_depth'][contents[0]] = [float(i) for i in contents[1:3]]


def get_convolutions(dataset, pdb_atom_names,
                     atom_label=('set', 'string')[0],
                     perform_checks=True,
                     v=5,
                     order=False,
                     return_type=['mask', 'idxs'][1],
                     absolute_torsion_period=True,
                     NB=('matrix',)[0],
                     fix_terminal=True,
                     fix_charmm_residues=True,
                     fix_slice_method=False,
                     fix_h=False,
                     alt_vdw=[],
                     permitivity=1.0
                     ):
    '''
    ##INPUTS##

    dataset:         one frame of a trajectory of shape [3, N]

    pdb_atom_names:  should be an array of shape [N,2]
                     pdb_atom_names[:,0] is the pdb_atom_names and
                     pdb_atom_names[:,1] is the residue names

    atom_label:      (default, 'set') deprecated and broken for anything other than 'set'

    perform_checks:  No longer works so has been removed

    v:               (default, 5) atom_selection version, bonds are determined by interatomic
                     distance, with v=2. v=1 shouldn't be used except in specific cirumstances. v=5 is using connectivity from amber parameters.

    order:           (bool, default false) are atoms ordered, I think I've fixed this so that
                     it shouldn't matter either way but keep as False.

    return_type:     (option now removed)


    ##OUTPUTS##
    convolution output shape N* will be N-(conv length -1)+padding

    bond_masks, b_equil, b_force: shape [number of convolutions, N*]

    bond_weights: shape[number of convolutions, conv_size]

    angle_masks, a_equil, a_force:  shape [number of convolutions,  conv_size]

    angle_weights: shape[number of convolutions, 2, conv_size]

    torsion_masks: shape[number of convolution, 3, conv_size]

    t_para:        shape[num of convs, N*, 4, max number torsion parameters ]

    tornsion_weigths: shape [number of convolutions, 3, conv_size]

    '''

    # get amber parameters
    (amber_atoms, atom_mass, atom_polarizability, bond_force, bond_equil,
     angle_force, angle_equil, torsion_factor, torsion_barrier, torsion_phase,
     torsion_period, improper_factor, improper_barrier, improper_phase,
     improper_period, other_parameters) = get_amber_parameters()
    if fix_terminal:
        pdb_atom_names[pdb_atom_names[:, 0] == 'OXT', 0] = 'O'
    if fix_charmm_residues:
        pdb_atom_names[pdb_atom_names[:, 1] == 'HSD', 1] = 'HID'
        pdb_atom_names[pdb_atom_names[:, 1] == 'HSE', 1] = 'HIE'
        for i in np.unique(pdb_atom_names[:, 2]):
            res_mask = pdb_atom_names[:, 2] == i
            if (pdb_atom_names[res_mask, 1] == 'HIS').all():  # if a HIS residue
                if (pdb_atom_names[res_mask, 0] == 'HD1').any() and (pdb_atom_names[res_mask, 0] == 'HE2').any():
                    pdb_atom_names[res_mask, 1] = 'HIP'
                elif (pdb_atom_names[res_mask, 0] == 'HD1').any():
                    pdb_atom_names[res_mask, 1] = 'HID'
                elif (pdb_atom_names[res_mask, 0] == 'HE2').any():
                    pdb_atom_names[res_mask, 1] = 'HIE'
        # if any HIS are remaining it does not matter which because the H is dealt with above
        pdb_atom_names[pdb_atom_names[:, 1] == 'HIS', 1] = 'HIE'
    if fix_h:
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'MET'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'MET'), 0] = 'HG3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'ASN'), 0] = 'HB3'
        pdb_atom_names[pdb_atom_names[:, 0] == 'HN', 0] = 'H'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HD2', 0] = 'HD21'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HD2', 0] = 'HD22'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HG2', 0] = 'HG21'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HG2', 0] = 'HG22'
        pdb_atom_names[pdb_atom_names[:, 0] == '3HG2', 0] = 'HG23'
        pdb_atom_names[pdb_atom_names[:, 0] == '3HG1', 0] = 'HG13'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HG1', 0] = 'HG11'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HG1', 0] = 'HG12'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HD1', 0] = 'HD11'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HD1', 0] = 'HD12'
        pdb_atom_names[pdb_atom_names[:, 0] == '3HD1', 0] = 'HD13'
        pdb_atom_names[pdb_atom_names[:, 0] == '3HD2', 0] = 'HD23'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HH1', 0] = 'HH11'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HH1', 0] = 'HH12'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HH2', 0] = 'HH21'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HH2', 0] = 'HH22'
        pdb_atom_names[pdb_atom_names[:, 0] == '1HE2', 0] = 'HE21'
        pdb_atom_names[pdb_atom_names[:, 0] == '2HE2', 0] = 'HE22'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG11', pdb_atom_names[:, 1] == 'ILE'), 0] = 'HG13'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'CD', pdb_atom_names[:, 1] == 'ILE'), 0] = 'CD1'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HD1', pdb_atom_names[:, 1] == 'ILE'), 0] = 'HD11'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HD2', pdb_atom_names[:, 1] == 'ILE'), 0] = 'HD12'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HD3', pdb_atom_names[:, 1] == 'ILE'), 0] = 'HD13'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'PHE'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'GLU'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'GLU'), 0] = 'HG3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'LEU'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'ARG'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'ARG'), 0] = 'HG3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HD1', pdb_atom_names[:, 1] == 'ARG'), 0] = 'HD3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'ASP'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HA1', pdb_atom_names[:, 1] == 'GLY'), 0] = 'HA3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'LYS'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'LYS'), 0] = 'HG3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HD1', pdb_atom_names[:, 1] == 'LYS'), 0] = 'HD3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HE1', pdb_atom_names[:, 1] == 'LYS'), 0] = 'HE3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'TYR'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'HIP'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'SER'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'SER'), 0] = 'HG'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'PRO'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'PRO'), 0] = 'HG3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HD1', pdb_atom_names[:, 1] == 'PRO'), 0] = 'HD3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'LEU'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'GLN'), 0] = 'HB3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HG1', pdb_atom_names[:, 1] == 'GLN'), 0] = 'HG3'
        pdb_atom_names[np.logical_and(pdb_atom_names[:, 0] == 'HB1', pdb_atom_names[:, 1] == 'TRP'), 0] = 'HB3'
    # writes termini as H because we haven't loaded in termini parameters
    atom_names = [[amber_atoms[res][atom], res, resid] if atom not in ['H2', 'H3'] else [amber_atoms[res]['H'], res, resid] for atom, res, resid in pdb_atom_names]
    p_atom_names = [[atom, res, resid] if atom not in ['H2', 'H3'] else ['H', res, resid] for atom, res, resid in pdb_atom_names]

    # atom_names = [[amber_atoms[res][atom],res] for atom, res, resid in pdb_atom_names ]
    atom_charges = [other_parameters['charge'][res][atom] for atom, res, resid in atom_names]
    if NB == 'matrix':
        equiv_t = other_parameters['equivalences']
        vdw_para = other_parameters['vdw_potential_well_depth']
        # switch these around so that values point to key
        equiv = {}
        for i in equiv_t.keys():
            j = equiv_t[i]
            for k in j:
                equiv[k] = i
        atom_R = torch.tensor([vdw_para[equiv.get(atom, atom)][0] for atom, res, resid in atom_names])  # radius
        atom_e = torch.tensor([vdw_para[equiv.get(atom, atom)][1] for atom, res, resid in atom_names])  # welldepth

    print('Determining bonds')
    version = v  # method of selecting bonded atoms
    N = dataset.shape[1]  # 145

    cmat = (torch.nn.functional.pdist((dataset).permute(1, 0))).cpu().numpy()
    if version == 1:
        bond_idxs = np.argpartition(cmat, (N - 1, N))
        # this will work for any non cyclic monomeric protein
        # that in mind will break if enough proline atoms to make a cycle are selected
        bond_idxs, u = bond_idxs[:N - 1], bond_idxs[N - 1]
        if cmat[u] - cmat[bond_idxs[-1]] < 0.25:
            dist_val = str(cmat[u] - cmat[bond_idxs[-1]])
            msg = ("WARNING: May not have correctly selected the bonded distances: value %s "
                   "should be roughly between 0.42 and 0.57 (>0.25)" % dist_val)
            raise Exception(msg)  # should be 0.42-0.57
            version += 1  # try version 2 instead
        mid = cmat[bond_idxs[-1]] + ((cmat[u] - cmat[bond_idxs[-1]]) / 2)  # mid point
        full_mask = (cmat < mid).astype('int8')
    if version == 2:
        full_mask = (cmat < (1.643 + 2.129) / 2).astype('int8')
        bond_idxs = np.where(full_mask)[0]  # for some reason returns tuple with one array
    if version == 3:
        if alt_vdw:
            vdw = torch.tensor(alt_vdw)
            max_bond_dist = (0.6 * (vdw.view(1, -1) + vdw.view(-1, 1)))
            cdist = torch.cdist(dataset.T, dataset.T)
            i, j = np.where((max_bond_dist > cdist).triu(diagonal=1).numpy())
            remove = np.where(np.abs(j - i) > 30)
            max_bond_dist[i[remove], j[remove]] = 0.0
            max_bond_dist = max_bond_dist.numpy()
        else:
            max_bond_dist = (0.6 * (atom_R.view(1, -1) + atom_R.view(-1, 1))).cpu().numpy()
        max_bond_dist = max_bond_dist[np.where(np.triu(np.ones((N, N)), k=1))]
        full_mask = np.greater(max_bond_dist, cmat)
        bond_idxs = np.where(full_mask)[0]  # for some reason returns tuple with one array
    if version == 4:
        # fix_hydrogens = [[atom, res, resid] for atom, res, resid in pdb_atom_names if atom in ['H2', 'H3']]
        connectivity = other_parameters['connectivity']
        bond_types = []
        bond_idxs = []
        # tracker = [[]]*N doesn't work because of mutability
        tracker = [[] for i in range(N)]
        current_resid = -9999
        current_atoms = []
        for i1, (atom1, res, resid) in enumerate(p_atom_names):
            assert atom1 in connectivity[res]
            for atom2, i2 in current_atoms:
                if resid != current_resid:  # and atom2 == 'C':
                    if atom2 != 'C':
                        continue
                elif not (atom2 in connectivity[res][atom1] and atom1 in connectivity[res][atom2]):
                    continue
                #   if not (atom2 in connectivity[res][atom1] and atom1 in connectivity[res][atom2]):
                #       if resid != current_resid and atom2 == 'C':
                #           current_resid = resid
                #           current_atoms = []
                #       else:
                #           continue
                if atom1 == 'N' and atom2 == 'CA':
                    continue
                tracker[i1].append(i2)
                tracker[i2].append(i1)
                if atom_label == 'set':
                    if order:
                        names = tuple(sorted((atom_names[i2][0], atom_names[i1][0])))
                    else:
                        names = tuple((atom_names[i2][0], atom_names[i1][0]))
                    bond_types.append(names)
                    bond_idxs.append([i2, i1])
            if resid != current_resid:  # and atom2 == 'C':
                current_resid = resid
                current_atoms = []
            current_atoms.append([atom1, i1])
    if version == 5:
        connectivity = other_parameters['connectivity']
        bond_types = []
        bond_idxs = []
        tracker = [[] for i in range(N)]
        current_resid = -9999
        current_atoms = []
        previous_atoms = []
        for i1, (atom1, res, resid) in enumerate(p_atom_names):
            assert atom1 in connectivity[res]
            if resid != current_resid:
                previous_atoms = deepcopy(current_atoms)
                current_atoms = []
                current_resid = resid
            if atom1 == 'N':
                for atom2, i2 in previous_atoms:
                    if atom2 == 'C':
                        tracker[i1].append(i2)
                        tracker[i2].append(i1)
                        bond_types.append(tuple((atom_names[i2][0], atom_names[i1][0])))
                        bond_idxs.append([i2, i1])

            for atom2, i2 in current_atoms:
                if atom2 in connectivity[res][atom1] and atom1 in connectivity[res][atom2]:
                    tracker[i1].append(i2)
                    tracker[i2].append(i1)
                    names = tuple((atom_names[i2][0], atom_names[i1][0]))
                    bond_types.append(names)
                    bond_idxs.append([i2, i1])
            current_atoms.append([atom1, i1])
    if version < 4:
        all_bond_idxs = np.sort(bond_idxs)

        bond_types = []
        bond_idxs = []
        tracker = [[]]  # this will keep track of some of the bonds to help work out the angles
        atom1 = 0
        atom2 = 1
        counter = 0  # index of the distance N,N+1
        for bond in all_bond_idxs:
            if bond < counter + (N - atom1 - 1):
                atom2 = atom1 + bond - counter + 1  # 0-0+1
                tracker[-1].append(atom2)
            while bond > counter + (N - atom1 - 2):
                counter += (N - atom1 - 1)
                atom1 += 1
                tracker.append([])
                if bond < counter + (N - atom1 - 1):
                    atom2 = atom1 + bond - counter + 1
                    tracker[-1].append(atom2)
            if atom_label == 'string':  # string of atom labels, doesn't handle Proline alternate ordering
                bond_types.append(atom_names[atom1][0] + '_' + atom_names[atom2][0])
                bond_idxs.append([atom1, atom2])
            elif atom_label == 'set':  # set of atom labels
                if order:
                    names = tuple(sorted((atom_names[atom1][0], atom_names[atom2][0])))
                else:
                    names = (atom_names[atom1][0], atom_names[atom2][0])
                bond_types.append(names)
                bond_idxs.append([atom1, atom2])

    while len(tracker) < N:
        tracker.append([])  # ensure so the next bit doesn't break by indexing N-1

    # Angles/1-3
    print('Determining angles')
    angle_types = []
    angle_idxs = []

    torsion_types = []
    torsion_idxs = []

    bond_14_idxs = []

    counter = 0
    # add missing bonds (each bond counted twice after but atom3>atom1 prevents duplicates later )
    if version < 4:
        for atom1, atom1_bonds in enumerate(deepcopy(tracker)):  # for _, [] in enum [[]]
            for atom2 in atom1_bonds:                            # for _ in []
                tracker[atom2].append(atom1)
    # find every angle and add it
    for atom1, atom1_bonds in enumerate(tracker):
        for atom2 in atom1_bonds:
            for atom3 in tracker[atom2]:
                if atom3 > atom1:  # each angle will only be counter once
                    if order:
                        sort13 = sorted([(atom_names[atom1][0], atom1), (atom_names[atom3][0], atom3)], key=lambda x: x[0])
                        names = tuple((sort13[0][0], atom_names[atom2][0], sort13[1][0]))

                        angle_types.append(names)
                        angle_idxs.append([sort13[0][1], atom2, sort13[1][1]])
                    else:
                        angle_types.append((atom_names[atom1][0], atom_names[atom2][0],
                                            atom_names[atom3][0]))
                        angle_idxs.append([atom1, atom2, atom3])
                if atom3 != atom1:
                    for atom4 in tracker[atom3]:
                        if atom4 > atom1 and atom2 != atom4:  # each torsion will be counter once
                            # torsions are done based on the 2 3 atoms, so sort 23
                            if order:
                                sort23 = sorted([(atom_names[atom2][0], atom2, atom_names[atom1][0], atom1),
                                                 (atom_names[atom3][0], atom3, atom_names[atom4][0], atom4)], key=lambda x: x[0])
                                names = tuple((sort23[0][2], sort23[0][0], sort23[1][0], sort23[1][2]))
                                torsion_types.append(names)
                                torsion_idxs.append([sort23[0][3], sort23[0][1], sort23[1][1], sort23[1][3]])
                            else:
                                torsion_types.append((atom_names[atom1][0], atom_names[atom2][0],
                                                      atom_names[atom3][0], atom_names[atom4][0]))
                                torsion_idxs.append([atom1, atom2, atom3, atom4])
                            bond_14_idxs.append([atom1, atom4])
    # currently have bond_types, angle_types, and torsion_typs + idxs
    bond_idxs = np.array(bond_idxs)
    angle_idxs = np.array(angle_idxs)
    torsion_idxs = np.array(torsion_idxs)
    bond_max_conv = (bond_idxs.max(axis=1) - bond_idxs.min(axis=1)).max() + 1
    if bond_max_conv < 3 and fix_slice_method:
        bond_max_conv = 3
    angle_max_conv = (angle_idxs.max(axis=1) - angle_idxs.min(axis=1)).max() + 1
    if angle_max_conv < 5 and fix_slice_method:
        angle_max_conv = 5
    torsion_max_conv = (torsion_idxs.max(axis=1) - torsion_idxs.min(axis=1)).max() + 1
    if torsion_max_conv < 7 and fix_slice_method:
        torsion_max_conv = 7
        # there is a problem where i accidentally index [padding-3] so if (len -4) < 3 we index -1 which breaks things
        # it shouldn't affect anything to say the max conv is greater than 6
    # this little bit just turns the 'types' list into equivalent parameters
    # key error if you don't have the parameter
    bond_para = np.array([[bond_equil[bond], bond_force[bond]] if bond in bond_equil
                          else [bond_equil[(bond[1], bond[0])], bond_force[(bond[1], bond[0])]]
                          for bond in bond_types])
    angle_para = np.array([[angle_equil[angle], angle_force[angle]] if angle in angle_equil
                           else [angle_equil[(angle[2], angle[1], angle[0])], angle_force[(angle[2], angle[1], angle[0])]]
                           for angle in angle_types])
    torsion_para = []
    t_unique = list(set(torsion_types))
    t_unique_para = {}
    max_para = 0
    for torsion in t_unique:
        torsion_b = (torsion[3], torsion[2], torsion[1], torsion[0])
        torsion_xx = ('X', torsion[2], torsion[1], 'X')
        torsion_xb = ('X', torsion[1], torsion[2], 'X')
        if torsion in torsion_barrier:
            max_para = max(max_para, len(torsion_barrier[torsion]))
            t_unique_para[torsion] = [torsion_factor[torsion], torsion_barrier[torsion],
                                      torsion_phase[torsion], torsion_period[torsion]]
        elif torsion_b in torsion_barrier:
            max_para = max(max_para, len(torsion_barrier[torsion_b]))
            t_unique_para[torsion] = [torsion_factor[torsion_b], torsion_barrier[torsion_b],
                                      torsion_phase[torsion_b], torsion_period[torsion_b]]
        elif torsion_xx in torsion_barrier:
            max_para = max(max_para, len(torsion_barrier[torsion_xx]))
            t_unique_para[torsion] = [torsion_factor[torsion_xx], torsion_barrier[torsion_xx],
                                      torsion_phase[torsion_xx], torsion_period[torsion_xx]]
        elif torsion_xb in torsion_barrier:
            max_para = max(max_para, len(torsion_barrier[torsion_xb]))
            t_unique_para[torsion] = [torsion_factor[torsion_xb], torsion_barrier[torsion_xb],
                                      torsion_phase[torsion_xb], torsion_period[torsion_xb]]
        else:
            print('ERROR: Torsion %s cannot be found in torsion_barrier and will not be included' % torsion)
    torsion_para = np.zeros((len(torsion_types), 4, max_para))
    # we do not want barrier/factor to return nan so set factor to 1 by default
    torsion_para[:, 0, :] = 1.0
    for i, torsion in enumerate(torsion_types):
        para = t_unique_para[torsion]
        torsion_para[i, :, :len(para[0])] = para
    # make phase positive
    if absolute_torsion_period:
        torsion_para[:, 3, :] = np.abs(torsion_para[:, 3, :])

    # bonds

    bond_masks = np.zeros((bond_max_conv - 1, N - (bond_max_conv - 1) + 2 * (bond_max_conv - 2)), dtype=bool)
    bond_conv = (bond_idxs.max(axis=1) - bond_idxs.min(axis=1)) - 1

    bond_weights = []
    b_equil = np.zeros(bond_masks.shape)
    b_force = np.zeros(bond_masks.shape)
    for i in range(bond_max_conv - 1):
        weight = [0.0] * bond_max_conv
        weight[0] = 1.0
        weight[i + 1] = -1.0
        bond_weights.append(weight)
        mask_index = bond_idxs.min(axis=1)[bond_conv == i] + bond_max_conv - 2
        bond_masks[i, mask_index] = True
        b_equil[i, mask_index] = bond_para[bond_conv == i, 0]
        b_force[i, mask_index] = bond_para[bond_conv == i, 1]

    # angles

    angle_conv = (angle_idxs - angle_idxs.min(axis=1).reshape(-1, 1))  # relative positions of atoms
    angle_conv = np.where((angle_conv[:, 0] < angle_conv[:, 2]).reshape(-1, 1), angle_conv, angle_conv[:, [2, 1, 0]])  # remove mirrors
    angle_unique = np.unique(angle_conv, axis=0)  # unique

    angle_masks = np.zeros((len(angle_unique), N - (angle_max_conv - 1) + 2 * (angle_max_conv - 3)), dtype=bool)
    angle_weights = []
    a_equil = np.zeros(angle_masks.shape)
    a_force = np.zeros(angle_masks.shape)
    for i, angle in enumerate(angle_unique):
        weight = [[0.0] * angle_max_conv, [0.0] * angle_max_conv]  # 2xsize
        weight[0][angle[0]] = 1.0
        weight[0][angle[1]] = -1.0
        weight[1][angle[1]] = -1.0
        weight[1][angle[2]] = 1.0
        angle_weights.append(weight)
        mask_index = angle_idxs.min(axis=1)[(angle_conv == angle).all(axis=1)] + angle_max_conv - 3
        a_equil[i, mask_index] = angle_para[(angle_conv == angle).all(axis=1), 0]
        a_force[i, mask_index] = angle_para[(angle_conv == angle).all(axis=1), 1]
        angle_masks[i, mask_index] = True

    # torsion

    torsion_conv = (torsion_idxs - torsion_idxs.min(axis=1).reshape(-1, 1))  # relative positions of atoms
    torsion_conv = np.where((torsion_conv[:, 0] < torsion_conv[:, 3]).reshape(-1, 1), torsion_conv, torsion_conv[:, [3, 2, 1, 0]])  # remove mirrors
    torsion_unique = np.unique(torsion_conv, axis=0)  # unique

    torsion_masks = np.zeros((len(torsion_unique), N - (torsion_max_conv - 1) + 2 * (torsion_max_conv - 4)), dtype=bool)
    torsion_weights = []
    # ts = torsion_masks.shape
    t_para = np.zeros((torsion_masks.shape[0], torsion_masks.shape[1],
                       torsion_para.shape[1], torsion_para.shape[2]))
    # we do not want barrier/factor to return nan so set factor to 1 by default
    t_para[:, :, 0, :] = 1.0
    for i, torsion in enumerate(torsion_unique):
        weight = [[0.0] * torsion_max_conv, [0.0] * torsion_max_conv, [0.0] * torsion_max_conv]
        weight[0][torsion[0]] = 1.0  # b1 = ri-rj
        weight[0][torsion[1]] = -1.0
        weight[1][torsion[1]] = 1.0  # b2 = rj-rk
        weight[1][torsion[2]] = -1.0
        weight[2][torsion[2]] = -1.0  # b3 = rl-rk
        weight[2][torsion[3]] = 1.0
        torsion_weights.append(weight)
        mask_index = torsion_idxs.min(axis=1)[(torsion_conv == torsion).all(axis=1)] + torsion_max_conv - 4
        torsion_masks[i, mask_index] = True
        t_para[i, mask_index] = torsion_para[(torsion_conv == torsion).all(axis=1)]

    if NB == 'matrix':
        # cdist is easier to work with than pdist, batch pdist was removed from torch and has not been readded as of writting this
        vdw_R = 0.5 * torch.cdist(atom_R.view(-1, 1), -atom_R.view(-1, 1)).triu(diagonal=1)
        vdw_e = (atom_e.view(1, -1) * atom_e.view(-1, 1)).triu(diagonal=1).sqrt()
        # set 1-2, and 1-3 distances to 0.0
        vdw_R[bond_idxs.T] = 0.0
        vdw_e[bond_idxs.T] = 0.0
        vdw_R[angle_idxs[:, (0, 2)].T] = 0.0
        vdw_e[angle_idxs[:, (0, 2)].T] = 0.0
        vdw_R[torsion_idxs[:, (0, 3)].T] = 0.0
        vdw_e[torsion_idxs[:, (0, 3)].T] = 0.0

        e_ = permitivity  # permitivity
        atom_charges = torch.tensor(atom_charges)
        q1q2 = (atom_charges.view(1, -1) * atom_charges.view(-1, 1) / e_).triu(diagonal=1)  # Aij=bi*bj
        q1q2[bond_idxs.T] = 0.0
        q1q2[angle_idxs[:, (0, 2)].T] = 0.0
        q1q2[torsion_idxs[:, (0, 3)].T] = 0.0

        # 1-4 are should be included but scaled
        bond_14_idxs = np.array(bond_14_idxs)
        vdw_14R = 0.5 * (atom_R[bond_14_idxs[:, 0]] + atom_R[bond_14_idxs[:, 1]])
        vdw_14e = (atom_e[bond_14_idxs[:, 0]] + atom_e[bond_14_idxs[:, 1]]).sqrt()
        q1q2_14 = (atom_charges[bond_14_idxs[:, 0]] * atom_charges[bond_14_idxs[:, 1]]) / e_

        return (bond_masks, b_equil, b_force, bond_weights,
                angle_masks, a_equil, a_force, angle_weights,
                torsion_masks, t_para, torsion_weights,
                vdw_R, vdw_e, vdw_14R, vdw_14e,
                q1q2, q1q2_14)

    return (bond_masks, b_equil, b_force, bond_weights,
            angle_masks, a_equil, a_force, angle_weights,
            torsion_masks, t_para, torsion_weights)


def get_conv_pad_res(dataset, pdb_atom_names,
                     absolute_torsion_period=True,
                     NB=('matrix',)[0],
                     fix_terminal=True,
                     fix_charmm_residues=True,
                     correct_1_4=True,
                     permitivity=1.0):
    '''
    ##INPUTS##

    dataset:         one frame of a trajectory of shape [3, N]

    pdb_atom_names:  should be an array of shape [N,2]
                     pdb_atom_names[:,0] is the pdb_atom_names and
                     pdb_atom_names[:,1] is the residue names

    '''

    if len(dataset.shape) != 3:
        raise Exception(f'967 dataset frame here should be of shape [R, M, 3] not {str(dataset.shape)}')
    if dataset.shape != pdb_atom_names.shape:
        raise Exception('969 dataset.shape != pdb_atom_names.shape')

    # get amber parameters
    (amber_atoms, atom_mass, atom_polarizability, bond_force, bond_equil,
     angle_force, angle_equil, torsion_factor, torsion_barrier, torsion_phase,
     torsion_period, improper_factor, improper_barrier, improper_phase,
     improper_period, other_parameters) = get_amber_parameters()

    R, M, D = dataset.shape  # N residues, Max atom per res, dimension D
    N = R * M
    # [R*M, 3], 3 = atom, res, resid
    pdb_atom_names = pdb_atom_names.reshape(-1, 3)
    # [R*M, 3], 3 = x, y, z
    dataset = dataset.reshape(-1, 3)

    if fix_terminal:  # fix atoms and residues not in amber parameters
        pdb_atom_names[pdb_atom_names[:, 0] == 'OXT', 0] = 'O'
    if fix_charmm_residues:
        pdb_atom_names[pdb_atom_names[:, 1] == 'HSD', 0] = 'HID'
        pdb_atom_names[pdb_atom_names[:, 1] == 'HSE', 0] = 'HIE'

    # pdb atoms -> amber atom names and residues
    padded_atom_names = np.array([[amber_atoms[res][atom], res, resid] if atom is not None else [atom, res, resid] for atom, res, resid in pdb_atom_names])
    # unpadded_atom_names = [[amber_atoms[res][atom],res, resid] for atom, res, resid in pdb_atom_names if atom is not None]
    padded_atom_charges = np.array([other_parameters['charge'][res][atom] if atom is not None else np.nan for atom, res, _ in padded_atom_names])
    if padded_atom_names.shape != dataset.shape:  # just a little check
        raise Exception('996 padded_atom_names!=dataset.shape')
    atom_names = padded_atom_names
    atom_charges = padded_atom_charges
    print('Determining bonds')

    connect = other_parameters['connectivity']
    # connectivity = [[]]*N # careful with mutability
    connectivity = [[] for i in range(N)]
    current_resid = -9999
    current_atoms = []
    for i1, (atom1, res, resid) in enumerate(padded_atom_names):
        if atom1 is None:
            continue
        if resid != current_resid:
            current_resid = resid
            current_atoms = []
        assert atom1 in connect[res]
        for atom2, i2 in current_atoms:
            if atom2 in connect[res][atom1] and atom1 in connect[res][atom2]:
                connectivity[i1].append(i2)
                connectivity[i2].append(i1)
#   cmat = torch.cdist(dataset, dataset) #[R*M,3 ]-> [R*M, R*M]
#   #1.643 was max bond distance in MurD test, 2.129 was the smallest nonbonded distance
#   #can't say what the best solution is but somewhere in the middle will probably be okay
#   all_bond_mask = (cmat<(1.643+2.1269)/2).triu(diagonal=1) # [R*M,R*M]
#   bond_idxs = all_bond_mask.nonzero() # [B x 2]
#   #name_set = set(atom_names[:,0])
#
#    connectivity = [[] for i in range(N)] # this will keep track of some of the bonds to help work out the angles
#    for i,j in bond_idxs:
#        connectivity[i].append(j)
#        connectivity[j].append(i)
#    ##################### Angles/1-3 #####################
    print('Determining angles')

    bond_idxs_ = []
    angle_idxs = []
    torsion_idxs = []
    bond_14_idxs = []

    bond_para = []
    angle_para = []
    torsion_para_ = []

    for atom1, atom2_list in enumerate(connectivity):
        for atom2 in atom2_list:
            a1, a2 = atom_names[atom1][0], atom_names[atom2][0]
            if atom1 < atom2:  # stops any pair of atoms being selected twice
                bond_idxs_.append([atom1, atom2])
                for b in [(a1, a2), (a2, a1)]:
                    if b in bond_equil:
                        bond_para.append([bond_equil[b], bond_force[b]])
                        break  # break prevents any bond from beind added twice
                else:
                    raise Exception('No associated bond parameter')

            for atom3 in connectivity[atom2]:
                a3 = atom_names[atom3][0]
                if atom3 > atom1:  # each angle will only be counter once
                    angle_idxs.append([atom1, atom2, atom3])
                    for a in [(a1, a2, a3), (a3, a2, a1)]:
                        if a in angle_equil:
                            angle_para.append([angle_equil[a], angle_force[a]])
                            break
                    else:
                        raise Exception('No associated angle parameter')
                if atom3 != atom1:  # don't go back to same atom
                    for atom4 in connectivity[atom3]:
                        if atom4 > atom1 and atom2 != atom4:
                            torsion_idxs.append([atom1, atom2, atom3, atom4])
                            bond_14_idxs.append([atom1, atom4])
                            a4 = atom_names[atom4][0]
                            for t in [(a1, a2, a3, a4), (a4, a3, a2, a1), ('X', a2, a3, 'X'), ('X', a3, a2, 'X')]:
                                if t in torsion_barrier:
                                    torsion_para_.append(torch.tensor([
                                        torsion_factor[t],
                                        torsion_barrier[t],
                                        torsion_phase[t],
                                        torsion_period[t]]))
                                    break  # each torsion only counter once
                            else:
                                raise Exception('No associated torsion parameter')

    bond_idxs = torch.as_tensor(bond_idxs_)
    angle_idxs = torch.tensor(angle_idxs)
    torsion_idxs = torch.tensor(torsion_idxs)
    bond_14_idxs = torch.tensor(bond_14_idxs)
    bond_para = torch.tensor(bond_para)
    angle_para = torch.tensor(angle_para)
    max_number_torsion_para = max([tf.shape[1] for tf in torsion_para_])
    torsion_para = torch.zeros(torsion_idxs.shape[0], 4, max_number_torsion_para)
    torsion_para[:, 0, :] = 1.0
    for i, tf in enumerate(torsion_para_):
        torsion_para[i, :, 0:tf.shape[1]] = tf
    if absolute_torsion_period:
        torsion_para[:, 3, :] = np.abs(torsion_para[:, 3, :])

    # Gather based potential
    # currently for data [B, R*M, 3] or [B, 3, N]
    aij0 = bond_idxs.reshape(-1, 2, 1).eq(angle_idxs[:, (0, 1)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    aij1 = bond_idxs.reshape(-1, 2, 1).eq(angle_idxs[:, (1, 0)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    ajk0 = bond_idxs.reshape(-1, 2, 1).eq(angle_idxs[:, (1, 2)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    ajk1 = bond_idxs.reshape(-1, 2, 1).eq(angle_idxs[:, (2, 1)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    ij_jk = torch.stack([torch.where((aij0 + aij1).T)[1], torch.where((ajk0 + ajk1).T)[1]])
    aij_ = aij1.float() - aij0.float()  # sign change needed for loss_function equation
    ajk_ = ajk0.float() - ajk1.float()
    angle_mask = torch.stack([aij_.sum(dim=0), ajk_.sum(dim=0)])

    # following are [N_bonds, N_torsions] arrays comparing if the ij or jk are the same
    ij0 = bond_idxs.reshape(-1, 2, 1).eq(torsion_idxs[:, (0, 1)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    ij1 = bond_idxs.reshape(-1, 2, 1).eq(torsion_idxs[:, (1, 0)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    jk0 = bond_idxs.reshape(-1, 2, 1).eq(torsion_idxs[:, (1, 2)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    jk1 = bond_idxs.reshape(-1, 2, 1).eq(torsion_idxs[:, (2, 1)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    kl0 = bond_idxs.reshape(-1, 2, 1).eq(torsion_idxs[:, (2, 3)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    kl1 = bond_idxs.reshape(-1, 2, 1).eq(torsion_idxs[:, (3, 2)].view(-1, 2, 1).permute(2, 1, 0)).all(dim=1)
    ij_jk_kl = torch.stack([torch.where((ij0 + ij1).T)[1],
                            torch.where((jk0 + jk1).T)[1],
                            torch.where((kl0 + kl1).T)[1]])
    ij_ = ij0.float() - ij1.float()
    jk_ = jk0.float() - jk1.float()
    kl_ = kl0.float() - kl1.float()
    torsion_mask = torch.stack([ij_.sum(dim=0), jk_.sum(dim=0), kl_.sum(dim=0)])

    # j-i i->j
    # i-j j->i reverse
    # k-j j->k
    # j-k k->j reverse
    # l-k k->l
    # k-l l->k reverse

    if NB == 'matrix':
        equiv_t = other_parameters['equivalences']
        vdw_para = other_parameters['vdw_potential_well_depth']
        # switch these around so that values point to key
        equiv = {}
        for i in equiv_t.keys():
            j = equiv_t[i]
            for k in j:
                equiv[k] = i
        atom_R = torch.tensor([vdw_para[equiv.get(i, i)][0] if i is not None else np.nan for i, j, k in atom_names])  # radius
        atom_e = torch.tensor([vdw_para[equiv.get(i, i)][1] if i is not None else np.nan for i, j, k in atom_names])  # welldepth
        # cdist is easier to work with than pdist, batch pdist doesn't seem to exist too
        vdw_R = 0.5 * torch.cdist(atom_R.view(-1, 1), -atom_R.view(-1, 1)).triu(diagonal=1)
        vdw_e = (atom_e.view(1, -1) * atom_e.view(-1, 1)).triu(diagonal=1).sqrt()
        # set 1-2, and 1-3 distances to 0.0
        vdw_R[list(bond_idxs.T)] = 0.0
        vdw_e[list(bond_idxs.T)] = 0.0
        vdw_R[list(angle_idxs[:, (0, 2)].T)] = 0.0
        vdw_e[list(angle_idxs[:, (0, 2)].T)] = 0.0
        if correct_1_4:
            # sum A/R**12 - B/R**6; A = e* (R**12); B = 2*e *(R**6)
            # therefore scale vdw by setting e /= 2.0
            # vdw_R[list(torsion_idxs[:,(0,3)].T)]/=2.0
            vdw_e[list(torsion_idxs[:, (0, 3)].T)] /= 2.0
        else:
            vdw_R[list(torsion_idxs[:, (0, 3)].T)] = 0.0
            vdw_e[list(torsion_idxs[:, (0, 3)].T)] = 0.0
        vdw_R[torch.isnan(vdw_R)] = 0.0
        vdw_e[torch.isnan(vdw_e)] = 0.0

        # partial charges are given as fragments of electron charge.
        # Can convert coulomb energy into kcal/mol by multiplying with 332.05.
        # therofore multiply q by sqrt(332.05)=18.22
        e_ = permitivity  # permittivity
        atom_charges = torch.tensor(atom_charges)
        q1q2 = (atom_charges.view(1, -1) * atom_charges.view(-1, 1) / e_).triu(diagonal=1)  # Aij=bi*bj
        q1q2[list(bond_idxs.T)] = 0.0
        q1q2[list(angle_idxs[:, (0, 2)].T)] = 0.0
        if correct_1_4:
            q1q2[list(torsion_idxs[:, (0, 3)].T)] /= 1.2
        else:
            q1q2[list(torsion_idxs[:, (0, 3)].T)] = 0.0
        # 1-4 are should be included but scaled
        return (bond_idxs, bond_para,
                angle_idxs, angle_para, angle_mask, ij_jk,
                torsion_idxs, torsion_para, torsion_mask, ij_jk_kl,
                vdw_R, vdw_e,
                q1q2)
    return (bond_idxs, bond_para,
            angle_idxs, angle_para, angle_mask, ij_jk,
            torsion_idxs, torsion_para, torsion_mask, ij_jk_kl)


if __name__ == '__main__':

    import sys
    sys.path.insert(0, os.path.abspath('../'))



#!/usr/bin/env python

import torch
import torch.nn as nn


class InformationBottleneckLoss(nn.Module):
    """
    Loss = reconstruction_error + beta * KL[q(z|x) || p(z)]

    Where p(z) is modeled as a mixture over representative_z (means/logvars),
    weighted by representative_weights(idle_input).
    """

    def __init__(
        self,
        beta: float = 1.0,
        eps: float = 1e-8,
    ):
        super().__init__()
        self.beta = beta
        self.eps = eps
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def log_p(
        self,
        z: torch.Tensor,                   # [B] or [B,1]
        rep_mean: torch.Tensor,            # [R]
        rep_logvar: torch.Tensor,          # [R]
        w: torch.Tensor,                   # [R] or [R,1]
        sum_up: bool = True
    ) -> torch.Tensor:
        """
        Compute log p(z) under the mixture prior.

        Args:
          z         (Tensor[B] or [B,1]): latent samples
          rep_mean  (Tensor[R])           : mixture means
          rep_logvar(Tensor[R])           : mixture log‐vars
          w         (Tensor[R] or [R,1])  : mixture weights
          sum_up    (bool)                : if True, returns [B] log‐density;
                                           if False, returns [B,R] component‐wise log‐probs
        Returns:
          Tensor[B] if sum_up else Tensor[B,R]
        """

        z_expand = z.unsqueeze(1)
        mu = rep_mean.unsqueeze(0)
        lv = rep_logvar.unsqueeze(0)

        representative_log_q = -0.5 * torch.sum(lv + torch.pow(z_expand - mu, 2) /
                                                torch.exp(lv), dim=2)

        if sum_up:
            log_p = torch.sum(torch.log(torch.exp(representative_log_q) @ w + self.eps), dim=1)
        else:
            log_p = torch.log(torch.exp(representative_log_q) * w.T + self.eps)

        return log_p

    def forward(
        self,
        data_targets: torch.Tensor,  # [B, C_out]
        outputs: torch.Tensor,      # [B, C_out], log‐probs
        z_sample: torch.Tensor,     # [B]
        z_mean: torch.Tensor,       # [B]
        z_logvar: torch.Tensor,     # [B]
        rep_mean: torch.Tensor,     # [R]
        rep_logvar: torch.Tensor,   # [R]
        w: torch.Tensor,            # [R] or [R,1]
        data_weights: torch.Tensor = None,
        sum_up: bool = True,
    ):
        """
        Computes:
          rec_err = E_q[−log p(x|z)]
          kld     = E_q[ log q(z|x) − log p(z) ]
        Returns:
          loss, rec_err (scalar), kl_term (scalar)
        """

        # --- RECONSTRUCTION ---
        # cross‐entropy per sample: [B]
        ce = torch.sum(-data_targets * outputs, dim=1)
        rec_err = torch.mean(ce * data_weights) if data_weights is not None else ce.mean()

        # --- KL TERM ---
        # log q(z|x): -½ ∑[logvar + (z−mean)² / exp(logvar)]
        log_q = -0.5 * (z_logvar + (z_sample - z_mean).pow(2).div(z_logvar.exp()))
        # log p(z): mixture prior
        log_p = self.log_p(z_sample, rep_mean, rep_logvar, w, sum_up=sum_up)

        # per‐sample KL
        kld = log_q - log_p
        kl_term = (kld * data_weights).mean() if data_weights is not None else kld.mean()

        loss = rec_err + self.beta * kl_term
        return loss, rec_err, kl_term
# class InformationBottleneckLoss(nn.Module):
#     """
#     Loss = reconstruction_error + beta * KL[q(z|x) || p(z)]
#     Where p(z) is a mixture over representative_z (means/logvars),
#     weighted by representative_weights(idle_input).
#     """
#     def __init__(self, beta: float = 1.0, eps: float = 1e-8):
#         super().__init__()
#         self.beta = beta
#         self.eps = eps
#         self.device = "cuda" if torch.cuda.is_available() else "cpu"

#     def log_p(
#         self,
#         z: torch.Tensor,                   # [B, n_cvs]
#         rep_mean: torch.Tensor,            # [k, n_cvs]
#         rep_logvar: torch.Tensor,          # [k, n_cvs]
#         w: torch.Tensor,                   # [k, 1]
#     ) -> torch.Tensor:
#         """
#         Compute log p(z) under the mixture prior.

#         Args:
#             z: (Tensor[B, n_cvs]) latent samples
#             rep_mean: (Tensor[k, n_cvs]) mixture means
#             rep_logvar: (Tensor[k, n_cvs]) mixture log-variances
#             w: (Tensor[k, 1]) mixture weights

#         Returns:
#             Tensor[B]
#         """
#         batch_size, n_cvs = z.shape
#         k = rep_mean.shape[0]

#         # Expand dimensions for broadcasting
#         z_expand = z.unsqueeze(1)  # [B, 1, n_cvs]
#         mu = rep_mean.unsqueeze(0)  # [1, k, n_cvs]
#         lv = rep_logvar.unsqueeze(0)  # [1, k, n_cvs]

#         var = torch.exp(lv)

#         # Log-probability per dimension per component
#         log_prob_per_dim = -0.5 * (math.log(2 * math.pi) + lv + ((z_expand - mu) ** 2) / var)  # [B, k, n_cvs]

#         # Sum over dimensions to get log_prob per component
#         log_prob_comp = log_prob_per_dim.sum(dim=2)  # [B, k]

#         # Add log-weights
#         log_w = torch.log(w + self.eps).squeeze(-1)  # [k]
#         log_prob_comp += log_w.unsqueeze(0)  # [B, k]

#         # Marginalize over components via logsumexp
#         log_p = torch.logsumexp(log_prob_comp, dim=1)  # [B]

#         return log_p

#     def forward(
#         self,
#         data_targets: torch.Tensor,  # [B, C_out]
#         outputs: torch.Tensor,       # [B, C_out], log-probs
#         z_sample: torch.Tensor,      # [B, n_cvs]
#         z_mean: torch.Tensor,        # [B, n_cvs]
#         z_logvar: torch.Tensor,      # [B, n_cvs]
#         rep_mean: torch.Tensor,      # [k, n_cvs]
#         rep_logvar: torch.Tensor,    # [k, n_cvs]
#         w: torch.Tensor,             # [k, 1]
#         data_weights: torch.Tensor = None,  # [B] or None
#     ):
#         """
#         Computes:
#             rec_err = E_q[-log p(x|z)]
#             kld = E_q[ log q(z|x) - log p(z) ]
#         Returns:
#             loss, rec_err (scalar), kl_term (scalar)
#         """
#         # Reconstruction error: cross-entropy per sample
#         ce = torch.sum(-data_targets * outputs, dim=1)  # [B]
#         rec_err = (ce * data_weights).mean() if data_weights is not None else ce.mean()

#         # KL term: log q(z|x) - log p(z)
#         # log q(z|x): full multivariate diagonal Gaussian log-prob
#         log_q_per_dim = -0.5 * (math.log(2 * math.pi) + z_logvar + ((z_sample - z_mean) ** 2) / z_logvar.exp())  # [B, n_cvs]
#         log_q = log_q_per_dim.sum(dim=1)  # [B]

#         # log p(z): mixture prior
#         log_p = self.log_p(z_sample, rep_mean, rep_logvar, w)  # [B]

#         # Per-sample KL
#         kld = log_q - log_p  # [B]
#         kl_term = (kld * data_weights).mean() if data_weights is not None else kld.mean()

#         loss = rec_err + self.beta * kl_term
#         return loss, rec_err, kl_term



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# feedforward.py
#
# from the mlcolvar repository
# -----------------------------------------------------------------------------

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Variational Autoencoder collective variable.
"""

__all__ = ["FeedForward"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

from typing import Optional, Union

import torch
import lightning
from .utils import get_activation, parse_nn_options

# =============================================================================
# STANDARD FEED FORWARD
# =============================================================================


class FeedForward(lightning.LightningModule):
    """Define a feedforward neural network given the list of layers.

    Optionally dropout and batchnorm can be applied (the order is activation -> dropout -> batchnorm).
    """

    def __init__(
        self,
        layers: list,
        activation: Union[str, list] = None,
        dropout: Optional[Union[float, list]] = None,
        batchnorm: Union[bool, list] = False,
        last_layer_activation: bool = False,
        **kwargs,
    ):
        """Constructor.

        Parameters
        ----------
        layers : list
            Number of neurons per layer.
        activation : string or list[str], optional
            Add activation function (options: relu, tanh, elu, linear). If a
            ``list``, this must have length ``len(layers)-1``, and ``activation[i]``
            controls whether to add the activation to the ``i``-layer.
        dropout : float or list[float], optional
            Add dropout with this probability after each layer. If a ``list``,
            this must have length ``len(layers)-1``, and ``dropout[i]`` specifies
            the the dropout probability for the ``i``-th layer.
        batchnorm : bool or list[bool], optional
            Add batchnorm after each layer. If a ``list``, this must have
            length ``len(layers)-1``, and ``batchnorm[i]`` controls whether to
            add the batchnorm to the ``i``-th layer.
        last_layer_activation : bool, optional
            If ``True`` and activation, dropout, and batchnorm are added also to
            the output layer when ``activation``, ``dropout``, or ``batchnorm``
            (i.e., they are not lists). Otherwise, the output layer will be linear.
            This option is ignored for the arguments among ``activation``, ``dropout``,
            and ``batchnorm`` that are passed as lists.
        **kwargs:
            Optional arguments passed to torch.nn.Module
        """
        super().__init__(**kwargs)

        # Parse layers
        if not isinstance(layers[0], int):
            raise TypeError("layers should be a list-type of integers.")

        # Parse options per each hidden layer
        n_layers = len(layers) - 1

        # -- activation
        activation_list = parse_nn_options(activation, n_layers, last_layer_activation)
        if isinstance(last_layer_activation, str):
            last_layer_activation = None
        # -- dropout
        dropout_list = parse_nn_options(dropout, n_layers, last_layer_activation)
        # -- batchnorm
        batchnorm_list = parse_nn_options(batchnorm, n_layers, last_layer_activation)

        # Create network
        modules = []
        for i in range(len(layers) - 1):
            modules.append(torch.nn.Linear(layers[i], layers[i + 1]))
            activ, drop, norm = activation_list[i], dropout_list[i], batchnorm_list[i]

            if activ is not None:
                modules.append(get_activation(activ))

            if drop is not None:
                modules.append(torch.nn.Dropout(p=drop))

            if norm:
                modules.append(torch.nn.BatchNorm1d(layers[i + 1]))

        # store model and attributes
        self.nn = torch.nn.Sequential(*modules)
        self.in_features = layers[0]
        self.out_features = layers[-1]

    # def extra_repr(self) -> str:
    #    repr = f"in_features={self.in_features}, out_features={self.out_features}"
    #    return repr

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.nn(x)



import os
import torch
import importlib
from typing import Dict, Any, Type, Optional
from biobb_pytorch.mdae.models import __all__ as AVAILABLE_MODELS
from biobb_pytorch.mdae.utils.model_utils import assert_valid_kwargs
from biobb_pytorch.mdae.utils.log_utils import get_size
from biobb_common.tools.file_utils import launchlogger
from biobb_common.tools import file_utils as fu
from biobb_common.generic.biobb_object import BiobbObject


class BuildModel(BiobbObject):
    """
    | biobb_pytorch BuildModel
    | Build a Molecular Dynamics AutoEncoder (MDAE) PyTorch model.
    | Builds a PyTorch autoencoder from the given properties.

    Args:
        input_stats_pt_path (str): Path to the input model statistics file. File type: input. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/ref_input_model.pt>`_. Accepted formats: pt (edam:format_2333).
        output_model_pth_path (str) (Optional): Path to save the model in .pth format. File type: output. `Sample file <https://github.com/bioexcel/biobb_pytorch/raw/master/biobb_pytorch/test/reference/mdae/output_model.pth>`_. Accepted formats: pth (edam:format_2333).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **model_type** (*str*) - ("AutoEncoder") Name of the model class to instantiate (must exist in biobb_pytorch.mdae.models).
            * **n_cvs** (*int*) - (1) Dimensionality of the latent space.
            * **encoder_layers** (*list*) - ([16]) List of integers representing the number of neurons in each encoder layer.
            * **decoder_layers** (*list*) - ([16]) List of integers representing the number of neurons in each decoder layer.
            * **options** (*dict*) - ({"norm_in": {"mode": "min_max"}}) Additional options (e.g. norm_in, optimizer, loss_function, device, etc.).

    Examples:
        This example shows how to use the BuildModel class to build a PyTorch autoencoder model::

            from biobb_pytorch.mdae.build_model import buildModel

            input_stats_pt_path = "input_stats.pt"
            output_model_pth_file = "model.pth"

            n_features = 128
            prop = {
                'model_type': 'AutoEncoder',
                'n_cvs': 10,
                'encoder_layers': [n_features, 64, 32],
                'decoder_layers': [32, 64, n_features],
                'options': {
                    'norm_in': {"mode": "min_max"},
                    'optimizer': {
                        'lr': 1e-4
                    }
                }
            }

            buildModel(input_stats_pt_path=input_stats_pt_path,
                       output_model_pth_path=None,
                       properties=prop)

    Info:
        * wrapped_software:
            * name: PyTorch
            * version: >=1.6.0
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
        self,
        input_stats_pt_path: str,
        output_model_pth_path: Optional[str] = None,
        properties: dict = None,
        **kwargs,
    ) -> None:

        properties = properties or {}

        super().__init__(properties)

        self.input_stats_pt_path = input_stats_pt_path
        self.output_model_pth_path = output_model_pth_path
        self.props = properties.copy()
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_stats_pt_path": input_stats_pt_path,
            },
            "out": {}
        }
        if output_model_pth_path:
            self.io_dict["out"]["output_model_pth_path"] = output_model_pth_path

        # build the per-feature arguments
        self.options: dict = properties.get("options", {})
        self.model_type: str = properties.get("model_type", "AutoEncoder")
        self.n_cvs: int = properties.get("n_cvs", 1)
        self.encoder_layers: list = properties.get("encoder_layers", [16])
        self.decoder_layers: list = properties.get("decoder_layers", [16])
        self.loss_function: Optional[dict] = properties.get("loss_function", None)
        self.device = self.options['device'] if 'device' in self.options else 'cpu'

        # load the input files
        self.stats = torch.load(self.io_dict['in']['input_stats_pt_path'],
                                weights_only=False)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

        self._validate_props()
        self.model = self._build_model()
        self.loss_fn = self._build_loss()

        # Store hyperparameters for reproducibility
        hparams = {
            'model_type': properties['model_type'],
            'n_cvs': properties['n_cvs'],
            'encoder_layers': properties['encoder_layers'],
            'decoder_layers': properties['decoder_layers'],
            'loss_function': self._hparams_loss_repr(),
            'options': {k: v for k, v in properties['options'].items() if k != 'loss_function'}
        }
        setattr(self.model, '_hparams', hparams)

        # Attach loss_fn and move model to device
        self.model.loss_fn = self.loss_fn
        self.model.to(self.device)

    def _validate_props(self) -> None:
        required = ['model_type', 'n_cvs', 'encoder_layers', 'decoder_layers', 'options']
        missing = [k for k in required if k not in self.props]
        if missing:
            raise KeyError(f"Missing required properties: {missing}")

        model_type = self.props['model_type']
        if model_type not in AVAILABLE_MODELS:
            raise ValueError(
                f"Unknown model_type '{model_type}'. Available: {AVAILABLE_MODELS}"
            )

    def _build_model(self) -> torch.nn.Module:
        module = importlib.import_module('biobb_pytorch.mdae.models')
        ModelClass: Type[torch.nn.Module] = getattr(module, self.props['model_type'])

        init_args = {
            'n_features': self.stats['shape'][1],
            'n_cvs': self.props['n_cvs'],
            'encoder_layers': self.props['encoder_layers'],
            'decoder_layers': self.props['decoder_layers'],
            'options': {k: v for k, v in self.props['options'].items() if k not in ['loss_function', 'norm_in']}
        }

        if 'norm_in' in self.props.get('options', {}):

            init_args['options']['norm_in'] = {
                'stats': self.stats,
                'mode': self.props['options']['norm_in'].get('mode')
            }

        assert_valid_kwargs(ModelClass, init_args, context="model init")

        return ModelClass(**init_args)

    def _build_loss(self) -> torch.nn.Module:
        loss_config = self.props['options'].get('loss_function')
        if loss_config and 'loss_type' in loss_config and loss_config['loss_type'] == 'PhysicsLoss':
            loss_config['stats'] = self.stats

        if not loss_config:
            # Use model's default
            return getattr(self.model, 'loss_fn', None)

        loss_type = loss_config.get('loss_type')
        if not loss_type:
            raise KeyError("'loss_type' must be specified in options['loss_function']")

        loss_module = importlib.import_module('biobb_pytorch.mdae.loss')
        LossClass = getattr(loss_module, loss_type)

        kwargs = {k: v for k, v in loss_config.items() if k != 'loss_type'}

        assert_valid_kwargs(LossClass, kwargs, context="loss init")
        try:
            return LossClass(**kwargs)
        except Exception:
            kwargs = {k: v for k, v in kwargs.items() if k != 'stats'}
            return LossClass(**kwargs)

    def _hparams_loss_repr(self) -> str:
        loss_config = self.props['options'].get('loss_function')
        if loss_config:
            name = loss_config.get('loss_type', '')
            args = [f"{k}={v}" for k, v in loss_config.items() if k not in ['loss_type', 'stats']]
            return f"{name}({', '.join(args)})"
        # fallback to model's representation
        return repr(getattr(self.model, 'loss_fn', ''))

    def save_weights(self, path: str) -> None:
        """Save model.state_dict() to the given path."""
        torch.save(self.model.state_dict(), path)

    @classmethod
    def load_weights(
        cls,
        props: Dict[str, Any],
        path: str
    ) -> 'BuildModel':
        """Instantiate from props and load state_dict from path."""
        inst = cls(props)
        state = torch.load(path, map_location=inst.device)
        inst.model.load_state_dict(state)
        inst.model.to(inst.device)
        return inst

    def save_full(self) -> None:
        """Serialize the full model object (including architecture)."""
        torch.save(self.model, self.output_model_pth_path)

    @staticmethod
    def load_full(path: str) -> torch.nn.Module:
        """Load a model serialized with save_full."""
        return torch.load(path, weights_only=False)

    @launchlogger
    def launch(self) -> int:
        """
        Execute the :class:`BuildModel` class and its `.launch()` method.

        Args:
            output_model_pth_path (str): Path where the model will be saved.
            properties (dict): Hyper‐parameters for model construction.
        """

        # Setup Biobb
        if self.check_restart():
            return 0

        self.stage_files()

        if self.output_model_pth_path:
            self.save_full()

        fu.log("## BioBB AutoEncoder Builder ##", self.out_log)
        fu.log("", self.out_log)
        fu.log("Hyperparameters:", self.out_log)
        fu.log("----------------", self.out_log)
        for key, value in self.model.__dict__.get('_hparams').items():
            if key == 'options':
                fu.log(f"{key}:", self.out_log)
                for sub_key, sub_value in value.items():
                    fu.log(f"   {sub_key}: {sub_value}", self.out_log)
            else:
                fu.log(f"{key}: {value}", self.out_log)
        fu.log("", self.out_log)

        fu.log("Model:", self.out_log)
        fu.log("------", self.out_log)

        for line in str(self.model).splitlines():
            fu.log(line, self.out_log)
        fu.log("", self.out_log)

        if self.output_model_pth_path:
            fu.log(f"Model saved in .pth format in "
                   f'{os.path.abspath(self.io_dict["out"]["output_model_pth_path"])}',
                   self.out_log,
                   )
            fu.log(f'File size: '
                   f'{get_size(self.io_dict["out"]["output_model_pth_path"])}',
                   self.out_log,
                   )

        # Copy files to host
        self.copy_to_host()

        # Remove temporal files
        self.remove_tmp_files()

        self.check_arguments(output_files_created=bool(self.output_model_pth_path), raise_exception=False)

        return 0


def buildModel(
    properties: dict,
    input_stats_pt_path: str,
    output_model_pth_path: Optional[str] = None,
    **kwargs,
) -> int:
    """Create the :class:`BuildModel <BuildModel.BuildModel>` class and
    execute the :meth:`launch() <BuildModel.BuildModel.launch>` method."""
    return BuildModel(**dict(locals())).launch()


buildModel.__doc__ = BuildModel.__doc__
main = BuildModel.get_main(buildModel, "Build a Molecular Dynamics AutoEncoder (MDAE) PyTorch model.")


if __name__ == "__main__":
    main()

# Example usage:

# n_features = torch.rand(100, 20)
# n_feat = n_features.shape[1]

# properties = {
#             'model_type': 'VariationalAutoEncoder',
#             'n_cvs': 10,
#             'encoder_layers': [n_feat, 64, 32],
#             'decoder_layers': [32, 64, n_feat],
#             'options': {
#                 'loss_function': {
#                     'loss_type': 'ELBOLoss',
#                     'beta': 1.0,
#                     'reconstruction': 'mse',
#                     'reduction': 'sum'},
#                 'optimizer': {
#                     'lr': 0.001
#                 }
#             }
#         }

# model_builder = BuildModel(properties)
# model_builder.save_full("test_model.pth")
# model = model_builder.load_full("test_model.pth")

# print()
# print("Hyperparameters:")
# print("----------------")
# for key, value in model._hparams.items():
#     print(f"{key}: {value}")
# print()
# print("Model:")
# print("------")
# print(model)



import os
import pandas as pd
from openmm import Platform
from openmm.app import ForceField, PDBFile, Simulation
from openmm.app import element as elem
import openmm
from openmm.app.forcefield import _createResidueSignature
from openmm.app.internal import compiled
from torchexposedintegratorplugin import TorchExposedIntegrator

import torch
import numpy as np
from copy import deepcopy

soft_xml_script = """\
<ForceField>
 <Script>
import openmm as mm
nb = mm.CustomNonbondedForce('C/((r/0.2)^4+1)')
nb.addGlobalParameter('C', 1.0)
sys.addForce(nb)
for i in range(sys.getNumParticles()):
    nb.addParticle([])
exclusions = set()
for bond in data.bonds:
    exclusions.add((min(bond.atom1, bond.atom2), max(bond.atom1, bond.atom2)))
for angle in data.angles:
    exclusions.add((min(angle[0], angle[2]), max(angle[0], angle[2])))
for a1, a2 in exclusions:
    nb.addExclusion(a1, a2)
 </Script>
</ForceField>
"""


class ModifiedForceField(ForceField):

    def __init__(self, *args, alternative_residue_names=None, **kwargs):
        '''
        Takes all `*args` and `**kwargs` of `openmm.app.ForceField`, plus an optional parameter described here.

        :param dict  alternative_residue_names: aliases for resnames, e.g., `{'HIS':'HIE'}`.
        '''
        super().__init__(*args, **kwargs)
        if isinstance(alternative_residue_names, dict):
            self._alternative_residue_names = alternative_residue_names
        else:
            self._alternative_residue_names = {'HIS': 'HIE'}

    def _getResidueTemplateMatches(self, res, bondedToAtom, templateSignatures=None, ignoreExternalBonds=False, ignoreExtraParticles=False):
        """
        Return the templates that match a residue, or None if none are found.

        :param res: Topology.Residue, the residue for which template matches are to be retrieved.
        :param bondedToAtom: list of set of int, bondedToAtom[i] is the set of atoms bonded to atom index i
        :returns: list with two elements [template, matches].
                  _TemplateData is the matching forcefield residue template, or None if no matches are found.
                  matches is a list specifying which atom of the template each atom of the residue corresponds to,
                  or None if it does not match the template.
        """
        template = None
        matches = None
        for matcher in self._templateMatchers:
            template = matcher(self, res, bondedToAtom, ignoreExternalBonds, ignoreExtraParticles)
            if template is not None:
                match = compiled.matchResidueToTemplate(res, template, bondedToAtom, ignoreExternalBonds, ignoreExtraParticles)
                if match is None:
                    raise ValueError('A custom template matcher returned a template for residue %d (%s), but it does not match the residue.' % (res.index, res.name))
                return [template, match]
        if templateSignatures is None:
            templateSignatures = self._templateSignatures
        signature = _createResidueSignature([atom.element for atom in res.atoms()])
        if signature in templateSignatures:
            allMatches = []
            for t in templateSignatures[signature]:
                match = compiled.matchResidueToTemplate(res, t, bondedToAtom, ignoreExternalBonds, ignoreExtraParticles)
                if match is not None:
                    allMatches.append((t, match))
            if len(allMatches) == 1:
                template = allMatches[0][0]
                matches = allMatches[0][1]
            elif len(allMatches) > 1:
                for i, (t, m) in enumerate(allMatches):
                    name = self._alternative_residue_names.get(res.name, res.name)
                    if name == t.name.split('-')[0]:
                        template = t
                        matches = m
                        return [template, matches]
                    elif 'N' + name == t.name.split('-')[0]:
                        print(f'{str(res)}, {res.index}, is a being set as a N terminal residue')
                        template = t
                        matches = m
                        return [template, matches]
                    elif 'C' + name == t.name.split('-')[0]:
                        print(f'{str(res)} is a being set as a C terminal residue')
                        template = t
                        matches = m
                        return [template, matches]
                print(f'multiple for {t.name}')
                # We found multiple matches. This is OK if and only if they assign identical types and parameters to all atoms.
                t1, m1 = allMatches[0]

                for t2, m2 in allMatches[1:]:
                    if not t1.areParametersIdentical(t2, m1, m2):
                        raise Exception('Multiple non-identical matching templates found for residue %d (%s): %s.' % (res.index + 1, res.name, ', '.join(match[0].name for match in allMatches)))
                template = allMatches[0][0]
                matches = allMatches[0][1]
        return [template, matches]


class OpenmmPluginScore():
    '''
    This will use the new OpenMM Plugin to calculate forces and energy. The intention is that this will be fast enough to be able to calculate forces and energy during training.
    N.B.: The current torchintegratorplugin only supports float on GPU and double on CPU.
    '''

    def __init__(self, mol=None, xml_file=['amber14-all.xml'], platform='CUDA', remove_NB=False,
                 alternative_residue_names=dict(HIS='HIE', HSE='HIE'), atoms=['CA', 'C', 'N', 'CB', 'O'],
                 soft=False):
        '''
        :param `biobox.Molecule` mol: if pldataloader is not given, then a biobox object will be taken from this parameter. If neither are given then an error  will be thrown.
        :param str xml_file: xml parameter file
        :param str platform: 'CUDA' or 'Reference'.
        :param bool remove_NB: if True remove NonbondedForce, CustomGBForce and CMMotionRemover, else just remove CustomGBForce
        :param dict alternative_residue_names: aliases for resnames, e.g., `{'HIS':'HIE'}`.
        :param atoms:
        :param soft:
        '''
        self.mol = mol
        for key, value in alternative_residue_names.items():
            # self.mol.data.loc[:,'resname'][self.mol.data['resname']==key]=value
            self.mol.data.loc[self.mol.data['resname'] == key, 'resname'] = value
            # self.mol.data.loc[lambda df: df['resname']==key, key]=value
        tmp_file = f'tmp{np.random.randint(1e10)}.pdb'
        self.atoms = atoms
        self.mol.write_pdb(tmp_file, split_struc=False)
        self.pdb = PDBFile(tmp_file)
        if soft:
            print('attempting soft forcefield')
            from pdbfixer import PDBFixer
            f = PDBFixer(tmp_file)
            self.forcefield = f._createForceField(self.pdb.topology, False)
            self.system = self.forcefield.createSystem(self.pdb.topology)
        else:
            if isinstance(xml_file, str):
                self.forcefield = ModifiedForceField(xml_file, alternative_residue_names=alternative_residue_names)
            elif len(xml_file) > 0:
                self.forcefield = ModifiedForceField(*xml_file, alternative_residue_names=alternative_residue_names)
            else:
                raise ValueError(f'xml_file: {xml_file} needs to be a str or a list of str')

            if atoms == 'no_hydrogen':
                self.ignore_hydrogen()
            else:
                self.atomselect(atoms)
            # save pdb and reload in modeller
            templates, unique_unmatched_residues = self.forcefield.generateTemplatesForUnmatchedResidues(self.pdb.topology)
            self.system = self.forcefield.createSystem(self.pdb.topology)
            if remove_NB:
                forces = self.system.getForces()
                for idx in reversed(range(len(forces))):
                    force = forces[idx]
                    if isinstance(force, (  # openmm.PeriodicTorsionForce,
                            openmm.CustomGBForce,
                            openmm.NonbondedForce,
                            openmm.CMMotionRemover)):
                        self.system.removeForce(idx)
            else:
                forces = self.system.getForces()
                for idx in reversed(range(len(forces))):
                    force = forces[idx]
                    if isinstance(force, openmm.CustomGBForce):
                        self.system.removeForce(idx)

        self.integrator = TorchExposedIntegrator()
        self.platform = Platform.getPlatformByName(platform)
        self.simulation = Simulation(self.pdb.topology, self.system, self.integrator, self.platform)
        if platform == 'CUDA':
            self.platform.setPropertyValue(self.simulation.context, 'Precision', 'single')
        self.n_particles = self.simulation.context.getSystem().getNumParticles()
        self.simulation.context.setPositions(self.pdb.positions)
        self.get_score = self.get_energy
        print(self.simulation.context.getState(getEnergy=True).getPotentialEnergy()._value)
        os.remove(tmp_file)

    def ignore_hydrogen(self):
        # ignore = ['ASH', 'LYN', 'GLH', 'HID', 'HIP', 'CYM', ]
        ignore = []
        for name, template in self.forcefield._templates.items():
            if name in ignore:
                continue
            patchData = ForceField._PatchData(name + '_remove_h', 1)

            for atom in template.atoms:
                if atom.element is elem.hydrogen:
                    if atom.name not in patchData.allAtomNames:
                        patchData.allAtomNames.add(atom.name)
                        atomDescription = ForceField._PatchAtomData(atom.name)
                        patchData.deletedAtoms.append(atomDescription)
                    else:
                        raise ValueError()
            for bond in template.bonds:
                atom1 = template.atoms[bond[0]]
                atom2 = template.atoms[bond[1]]
                if atom1.element is elem.hydrogen or atom2.element is elem.hydrogen:
                    a1 = ForceField._PatchAtomData(atom1.name)
                    a2 = ForceField._PatchAtomData(atom2.name)
                    patchData.deletedBonds.append((a1, a2))
            self.forcefield.registerTemplatePatch(name, name + '_remove_h', 0)
            self.forcefield.registerPatch(patchData)

    def atomselect(self, atoms):
        atoms = deepcopy(atoms)
        if 'OT2' in atoms:
            atoms.append('OXT')
        if 'OT1' in atoms:
            atoms.append('OXT')

        for name, template in self.forcefield._templates.items():
            patchData = ForceField._PatchData(name + '_leave_only_' + '_'.join(atoms), 1)

            for atom in template.atoms:
                if atom.name not in atoms:
                    if atom.name not in patchData.allAtomNames:
                        patchData.allAtomNames.add(atom.name)
                        atomDescription = ForceField._PatchAtomData(atom.name)
                        patchData.deletedAtoms.append(atomDescription)
                    else:
                        raise ValueError()

            for bond in template.bonds:
                atom1 = template.atoms[bond[0]]
                atom2 = template.atoms[bond[1]]
                if atom1.name not in atoms or atom2.name not in atoms:
                    a1 = ForceField._PatchAtomData(atom1.name)
                    a2 = ForceField._PatchAtomData(atom2.name)
                    patchData.deletedBonds.append((a1, a2))
            self.forcefield.registerTemplatePatch(name, name + '_leave_only_' + '_'.join(atoms), 0)
            self.forcefield.registerPatch(patchData)

    def get_energy(self, pos_ptr, force_ptr, energy_ptr, n_particles, batch_size):
        '''
        :param pos_ptr: tensor.data_ptr()
        :param force_ptr: tensor.data_ptr()
        :param energy_ptr: tensor.data_ptr()
        :param int n_particles: number of particles
        :param int batch_size: batch size
        '''
        assert n_particles == self.n_particles
        torch.cuda.synchronize()
        self.integrator.torchMultiStructureE(pos_ptr, force_ptr, energy_ptr, n_particles, batch_size)
        return True

    def execute(self, x):
        '''
        :param `torch.Tensor` x: shape [b, N, 3]. dtype=float. device = gpu
        '''
        force = torch.zeros_like(x)
        energy = torch.zeros(x.shape[0], device=torch.device('cpu'), dtype=torch.double)
        self.get_energy(x.data_ptr(), force.data_ptr(), energy.data_ptr(), x.shape[1], x.shape[0])
        return force, energy


class OpenmmTorchEnergyMinimizer(OpenmmPluginScore):

    def minimize(self, x, maxIterations=10, threshold=10000):
        minimized_x = torch.empty_like(x)
        for i, s in enumerate(x.unsqueeze(1)):
            h = 0.01
            force, energy = self.execute(s)
            abs_max = 1 / (force.abs().max())
            for j in range(maxIterations):
                new_s = s - force * abs_max * h
                new_force, new_energy = self.execute(new_s)
                if new_energy < energy:
                    s, energy, force = new_s, new_energy, new_force
                    if energy < threshold:
                        break
                    h *= 1.2

                else:
                    h *= 0.2
            minimized_x[i] = s
        return minimized_x


class OpenMMPluginScoreSoftForceField(OpenmmPluginScore):

    def __init__(self, mol=None, platform='CUDA', atoms=['CA', 'C', 'N', 'CB', 'O']):
        self.mol = mol
        tmp_file = 'tmp.pdb'
        self.atoms = atoms
        self.mol.write_pdb(tmp_file, split_struc=False)
        self.pdb = PDBFile(tmp_file)
        from pdbfixer import PDBFixer
        f = PDBFixer(tmp_file)
        self.forcefield = f._createForceField(self.pdb.topology)
        self.system = self.forcefield.createSystem(self.pdb.topology)
        self.integrator = TorchExposedIntegrator()
        self.platform = Platform.getPlatformByName(platform)
        self.simulation = Simulation(self.pdb.topology, self.system, self.integrator, self.platform)
        if platform == 'CUDA':
            self.platform.setPropertyValue(self.simulation.context, 'Precision', 'single')
        self.n_particles = self.simulation.context.getSystem().getNumParticles()
        self.simulation.context.setPositions(self.pdb.positions)
        self.get_score = self.get_energy
        print(self.simulation.context.getState(getEnergy=True).getPotentialEnergy()._value)


class openmm_energy_function(torch.autograd.Function):

    @staticmethod
    def forward(ctx, plugin, x):
        '''
        :param plugin: OpenmmPluginScore instance
        :param `torch.Tensor` x: dtype = float, shape = [B, N, 3], device = any
        :returns: energy tensor, dtype = float, shape = [B], device  = any
        '''
        if x.device == torch.device('cpu'):
            force = np.zeros(x.shape)
            energy = np.zeros(x.shape[0])
            for i, t in enumerate(x):
                plugin.simulation.context.setPositions(t.numpy())
                state = plugin.simulation.context.getState(getForces=True, getEnergy=True)
                force[i] = state.getForces(asNumpy=True)
                energy[i] = state.getPotentialEnergy()._value
            force = torch.tensor(force).float()
            energy = torch.tensor(energy).float()
        else:
            # torch.cuda.synchronize(x.device)
            force, energy = plugin.execute(x)
            # torch.cuda.synchronize(x.device)
        ctx.save_for_backward(force)
        energy = energy.float().to(x.device)
        return energy

    @staticmethod
    def backward(ctx, grad_output):
        force = ctx.saved_tensors[0]  # force shape [B, N, 3]
        # embed(header='23 openmm_loss_function')
        return None, -force * grad_output.view(-1, 1, 1)


class openmm_clamped_energy_function(torch.autograd.Function):

    @staticmethod
    def forward(ctx, plugin, x, clamp):
        '''
        :param plugin: OpenmmPluginScore instance
        :param `torch.Tensor` x: dtype = float, shape = [B, N, 3], device = cuda
        :returns: energy tensor, dtype = double, shape = [B], device CPU
        '''
        if x.device == torch.device('cpu'):
            force = np.zeros(x.shape)
            energy = np.zeros(x.shape[0])
            for i, t in enumerate(x):
                plugin.simulation.context.setPositions(t.numpy())
                state = plugin.simulation.context.getState(getForces=True, getEnergy=True)
                force[i] = state.getForces(asNumpy=True)
                energy[i] = state.getPotentialEnergy()._value
            force = torch.tensor(force).float()
            energy = torch.tensor(energy).float()
        else:
            force, energy = plugin.execute(x)

        force = torch.clamp(force, **clamp)
        ctx.save_for_backward(force)
        energy = energy.float().to(x.device)
        return energy

    @staticmethod
    def backward(ctx, grad_output):
        force = ctx.saved_tensors[0]
        return None, -force * grad_output.view(-1, 1, 1), None


class openmm_energy(torch.nn.Module):

    def __init__(self, mol, std, clamp=None, **kwargs):
        super().__init__()
        self.openmmplugin = OpenmmPluginScore(mol, **kwargs)
        self.std = std / 10
        self.clamp = clamp
        if self.clamp is not None:
            self.forward = self._clamp_forward
        else:
            self.forward = self._forward

    def _forward(self, x):
        '''
        :param `torch.Tensor` x: dtype=torch.float, device=CUDA, shape B, 3, N
        :returns: torch energy tensor dtype should be float and on same device as x
        '''
        _x = (x * self.std).permute(0, 2, 1).contiguous()
        energy = openmm_energy_function.apply(self.openmmplugin, _x)
        return energy

    def _clamp_forward(self, x):
        '''
        :param `torch.Tensor` x: dtype=torch.float, device=CUDA, shape B, 3, N
        :returns: torch energy tensor dtype should be float and on same device as x
        '''
        _x = (x * self.std).permute(0, 2, 1).contiguous()
        energy = openmm_clamped_energy_function.apply(self.openmmplugin, _x, self.clamp)
        return energy


class openmm_energy_setup():

    def __init__(self, topology):

        self.top = topology

    def get_atominfo(self):
        """
        Get the atom information from the topology.
        """
        atom_info = []
        for i in self.top.topology.atoms:
            atom_info.append([i.name, i.residue.name, i.residue.index + 1])
        return np.array(atom_info, dtype=object)

    def mol_dataframe(self):
        """
        Get the atom information from the topology and convert it to a pandas DataFrame.
        """
        return MolDataFrame.from_stats(self.top)


class MolDataFrame(pd.DataFrame):
    """
    Subclass of pandas.DataFrame specialized for molecular data.
    Stores atomic information plus coordinates, and can write itself to PDB.
    """
    @property
    def data(self):
        return self

    @property
    def _constructor(self):
        return MolDataFrame

    @classmethod
    def from_trajectory(cls, traj, frame=0):
        """
        Build MolDataFrame from an mdtraj.Trajectory `traj`.
        - `frame`: index of the frame whose coords to use (default 0).
        Adds x, y, z columns.
        """
        # Atomic radii map (Å)
        radius_dict = {"N": 1.55, "C": 1.70, "O": 1.52,
                       "S": 1.80, "H": 1.20}

        def idx_to_letters(idx):
            letters = ''
            while idx >= 0:
                letters = chr(idx % 26 + ord('A')) + letters
                idx = idx // 26 - 1
            return letters

        rows = []
        coords = traj.xyz[frame]  # shape (n_atoms,3)
        for atom in traj.topology.atoms:
            i = atom.index
            # occupancy and B-factor
            occ = getattr(traj, 'occupancies', [[None]])[0][i]
            beta = getattr(traj, 'bfactors', [[None]])[0][i]
            occ = 1.0 if occ is None else occ
            beta = 0.0 if beta is None else beta

            name = atom.name
            element = atom.element.symbol if atom.element else None
            radius = radius_dict.get(name,
                                     radius_dict.get(element, None))
            charge = 0.0

            # chain as letters
            cid = idx_to_letters(atom.residue.chain.index)

            x, y, z = coords[i]
            rows.append({
                'atom': 'ATOM',
                'index': i,
                'name': name,
                'resname': atom.residue.name,
                'chain': cid,
                'resid': atom.residue.index + 1,
                'occupancy': occ,
                'beta': beta,
                'atomtype': element,
                'radius': radius,
                'charge': charge,
                'x': x,
                'y': y,
                'z': z,
            })

        cols = ['atom', 'index', 'name', 'resname', 'chain', 'resid',
                'occupancy', 'beta', 'atomtype', 'radius', 'charge',
                'x', 'y', 'z']
        return cls(rows, columns=cols)

    def write_pdb(self, filename):
        """
        Write this MolDataFrame (with x,y,z) to a PDB file.
        """
        fmt = (
            "ATOM  {index:5d} {name:^4s}{resname:>3s} {chain:1s}"
            "{resid:4d}    {x:8.3f}{y:8.3f}{z:8.3f}"
            "{occupancy:6.2f}{beta:6.2f}           {atomtype:>2s}\n"
        )
        with open(filename, 'w') as f:
            for row in self.to_dict('records'):
                f.write(fmt.format(**row))
            f.write("TER\nEND\n")
